from contextlib import ExitStack

import torch
import torch.nn.functional as F
from torch import nn

from ritm_annotation.model import ops

from .basic_blocks import SeparableConv2d
from .resnet import ResNetBackbone


class DeepLabV3Plus(nn.Module):
    def __init__(
        self,
        backbone="resnet50",
        norm_layer=nn.BatchNorm2d,
        backbone_norm_layer=None,
        ch=256,
        project_dropout=0.5,
        inference_mode=False,
        **kwargs
    ):
        super(DeepLabV3Plus, self).__init__()
        if backbone_norm_layer is None:
            backbone_norm_layer = norm_layer

        self.backbone_name = backbone
        self.norm_layer = norm_layer
        self.backbone_norm_layer = backbone_norm_layer
        self.inference_mode = False
        self.ch = ch
        self.aspp_in_channels = 2048
        self.skip_project_in_channels = 256  # layer 1 out_channels

        self._kwargs = kwargs
        if backbone == "resnet34":
            self.aspp_in_channels = 512
            self.skip_project_in_channels = 64

        self.backbone = ResNetBackbone(
            backbone=self.backbone_name,
            pretrained_base=False,
            norm_layer=self.backbone_norm_layer,
            **kwargs
        )

        self.head = _DeepLabHead(
            in_channels=ch + 32,
            mid_channels=ch,
            out_channels=ch,
            norm_layer=self.norm_layer,
        )
        self.skip_project = _SkipProject(
            self.skip_project_in_channels, 32, norm_layer=self.norm_layer
        )
        self.aspp = _ASPP(
            in_channels=self.aspp_in_channels,
            atrous_rates=[12, 24, 36],
            out_channels=ch,
            project_dropout=project_dropout,
            norm_layer=self.norm_layer,
        )

        if inference_mode:
            self.set_prediction_mode()

    def load_pretrained_weights(self):
        pretrained = ResNetBackbone(
            backbone=self.backbone_name,
            pretrained_base=True,
            norm_layer=self.backbone_norm_layer,
            **self._kwargs
        )
        backbone_state_dict = self.backbone.state_dict()
        pretrained_state_dict = pretrained.state_dict()

        backbone_state_dict.update(pretrained_state_dict)
        self.backbone.load_state_dict(backbone_state_dict)

        if self.inference_mode:
            for param in self.backbone.parameters():
                param.requires_grad = False

    def set_prediction_mode(self):
        self.inference_mode = True
        self.eval()

    def forward(self, x, additional_features=None):
        with ExitStack() as stack:
            if self.inference_mode:
                stack.enter_context(torch.no_grad())

            c1, _, c3, c4 = self.backbone(x, additional_features)
            c1 = self.skip_project(c1)

            x = self.aspp(c4)
            x = F.interpolate(
                x, c1.size()[2:], mode="bilinear", align_corners=True
            )
            x = torch.cat((x, c1), dim=1)
            x = self.head(x)

        return (x,)


class _SkipProject(nn.Module):
    def __init__(self, in_channels, out_channels, norm_layer=nn.BatchNorm2d):
        super(_SkipProject, self).__init__()
        _activation = ops.select_activation_function("relu")

        self.skip_project = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
            norm_layer(out_channels),
            _activation(),
        )

    def forward(self, x):
        return self.skip_project(x)


class _DeepLabHead(nn.Module):
    def __init__(
        self,
        out_channels,
        in_channels,
        mid_channels=256,
        norm_layer=nn.BatchNorm2d,
    ):
        super(_DeepLabHead, self).__init__()

        self.block = nn.Sequential(
            SeparableConv2d(
                in_channels=in_channels,
                out_channels=mid_channels,
                dw_kernel=3,
                dw_padding=1,
                activation="relu",
                norm_layer=norm_layer,
            ),
            SeparableConv2d(
                in_channels=mid_channels,
                out_channels=mid_channels,
                dw_kernel=3,
                dw_padding=1,
                activation="relu",
                norm_layer=norm_layer,
            ),
            nn.Conv2d(
                in_channels=mid_channels,
                out_channels=out_channels,
                kernel_size=1,
            ),
        )

    def forward(self, x):
        return self.block(x)


class _ASPP(nn.Module):
    def __init__(
        self,
        in_channels,
        atrous_rates,
        out_channels=256,
        project_dropout=0.5,
        norm_layer=nn.BatchNorm2d,
    ):
        super(_ASPP, self).__init__()

        b0 = nn.Sequential(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                bias=False,
            ),
            norm_layer(out_channels),
            nn.ReLU(),
        )

        rate1, rate2, rate3 = tuple(atrous_rates)
        b1 = _ASPPConv(in_channels, out_channels, rate1, norm_layer)
        b2 = _ASPPConv(in_channels, out_channels, rate2, norm_layer)
        b3 = _ASPPConv(in_channels, out_channels, rate3, norm_layer)
        b4 = _AsppPooling(in_channels, out_channels, norm_layer=norm_layer)

        self.concurent = nn.ModuleList([b0, b1, b2, b3, b4])

        project = [
            nn.Conv2d(
                in_channels=5 * out_channels,
                out_channels=out_channels,
                kernel_size=1,
                bias=False,
            ),
            norm_layer(out_channels),
            nn.ReLU(),
        ]
        if project_dropout > 0:
            project.append(nn.Dropout(project_dropout))
        self.project = nn.Sequential(*project)

    def forward(self, x):
        x = torch.cat([block(x) for block in self.concurent], dim=1)

        return self.project(x)


class _AsppPooling(nn.Module):
    def __init__(self, in_channels, out_channels, norm_layer):
        super(_AsppPooling, self).__init__()

        self.gap = nn.Sequential(
            nn.AdaptiveAvgPool2d((1, 1)),
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                bias=False,
            ),
            norm_layer(out_channels),
            nn.ReLU(),
        )

    def forward(self, x):
        pool = self.gap(x)
        return F.interpolate(
            pool, x.size()[2:], mode="bilinear", align_corners=True
        )


def _ASPPConv(in_channels, out_channels, atrous_rate, norm_layer):
    block = nn.Sequential(
        nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            padding=atrous_rate,
            dilation=atrous_rate,
            bias=False,
        ),
        norm_layer(out_channels),
        nn.ReLU(),
    )

    return block



import logging
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from pathlib import Path

import cv2
from tqdm import tqdm

COMMAND_DESCRIPTION = _(
    # noqa: E501
    "Looks for common problems in datasets in mask form that are used for finetuning and generated by the annotator"
)

logger = logging.getLogger(__name__)


def command(parser):
    parser.add_argument(
        "input",
        type=Path,
        help=_("Folder where the ground truths/masks are stored"),
    )
    parser.add_argument(
        "-i",
        "--images",
        dest="images",
        type=Path,
        help=_("Folder where the dataset images are stored"),
    )
    parser.add_argument(
        "-j",
        "--jobs",
        dest="jobs",
        type=int,
        default=cpu_count(),
        help=_("How many concurrent checks"),
    )

    def handle(args):
        assert args.input.is_dir(), _("Dataset must be a directory")
        assert args.images is None or args.images.is_dir(), _(
            "Invalid image directory"
        )
        if args.images is not None:
            logger.info(_("Using images with masks!"))

        def handle_one(item):
            if not item.is_dir():
                logger.warning(_("'{item}': Dataset noise").format(item=item))
                return
            item_img = None
            if args.images is not None:
                image_file = args.images / item.name
                if not image_file.exists():
                    logger.warning(
                        _("'{image_file}': Image file doesn't exist").format(
                            image_file=image_file
                        )
                    )
                else:
                    item_img = cv2.imread(str(image_file))
                    if item_img is None:
                        logger.error(
                            _("'{item_img}': Invalid image").format(
                                item_img=item_img
                            )
                        )
                    else:
                        item_img = cv2.cvtColor(item_img, cv2.COLOR_BGR2RGB)
            for mask in item.iterdir():
                mask_img = cv2.imread(str(mask), 0)
                if mask_img is None:
                    logger.error(_("'{mask}': Invalid mask").format(mask=mask))
                    continue
                if item_img is not None:
                    (wi, hi, di) = item_img.shape
                    (wm, hm) = mask_img.shape
                    if wi != wm:
                        logger.error(
                            _(
                                "'{mask}': First dimension doesn't match for image and mask"
                            ).format(
                                mask=mask
                            )  # noqa:E501
                        )
                    if hi != hm:
                        logger.error(
                            _(
                                "'{mask}': Second dimension doesn't match for image and mask"
                            ).format(
                                mask=mask
                            )  # noqa:E501
                        )

        items = list(args.input.iterdir())
        with ThreadPoolExecutor(max_workers=args.jobs) as executor:
            ops = tqdm(
                executor.map(handle_one, items, chunksize=8), total=len(items)
            )
            for item in ops:
                pass

    return handle



import torch
import torch._utils
import torch.nn as nn
import torch.nn.functional as F


class SpatialGather_Module(nn.Module):
    """
    Aggregate the context features according to the initial
    predicted probability distribution.
    Employ the soft-weighted method to aggregate the context.
    """

    def __init__(self, cls_num=0, scale=1):
        super(SpatialGather_Module, self).__init__()
        self.cls_num = cls_num
        self.scale = scale

    def forward(self, feats, probs):
        batch_size, c, _, _ = (
            probs.size(0),
            probs.size(1),
            probs.size(2),
            probs.size(3),
        )
        probs = probs.view(batch_size, c, -1)
        feats = feats.view(batch_size, feats.size(1), -1)
        feats = feats.permute(0, 2, 1)  # batch x hw x c
        probs = F.softmax(self.scale * probs, dim=2)  # batch x k x hw
        ocr_context = (
            torch.matmul(probs, feats).permute(0, 2, 1).unsqueeze(3)
        )  # batch x k x c
        return ocr_context


class SpatialOCR_Module(nn.Module):
    """
    Implementation of the OCR module:
    We aggregate the global object representation to update the representation for each pixel.
    """  # noqa: E501

    def __init__(
        self,
        in_channels,
        key_channels,
        out_channels,
        scale=1,
        dropout=0.1,
        norm_layer=nn.BatchNorm2d,
        align_corners=True,
    ):
        super(SpatialOCR_Module, self).__init__()
        self.object_context_block = ObjectAttentionBlock2D(
            in_channels, key_channels, scale, norm_layer, align_corners
        )
        _in_channels = 2 * in_channels

        self.conv_bn_dropout = nn.Sequential(
            nn.Conv2d(
                _in_channels,
                out_channels,
                kernel_size=1,
                padding=0,
                bias=False,
            ),
            nn.Sequential(norm_layer(out_channels), nn.ReLU(inplace=True)),
            nn.Dropout2d(dropout),
        )

    def forward(self, feats, proxy_feats):
        context = self.object_context_block(feats, proxy_feats)

        output = self.conv_bn_dropout(torch.cat([context, feats], 1))

        return output


class ObjectAttentionBlock2D(nn.Module):
    """
    The basic implementation for object context block
    Input:
        N X C X H X W
    Parameters:
        in_channels       : the dimension of the input feature map
        key_channels      : the dimension after the key/query transform
        scale             : choose the scale to downsample the input feature maps (save memory cost)
        bn_type           : specify the bn type
    Return:
        N X C X H X W
    """  # noqa: E501

    def __init__(
        self,
        in_channels,
        key_channels,
        scale=1,
        norm_layer=nn.BatchNorm2d,
        align_corners=True,
    ):
        super(ObjectAttentionBlock2D, self).__init__()
        self.scale = scale
        self.in_channels = in_channels
        self.key_channels = key_channels
        self.align_corners = align_corners

        self.pool = nn.MaxPool2d(kernel_size=(scale, scale))
        self.f_pixel = nn.Sequential(
            nn.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.key_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.Sequential(
                norm_layer(self.key_channels), nn.ReLU(inplace=True)
            ),
            nn.Conv2d(
                in_channels=self.key_channels,
                out_channels=self.key_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.Sequential(
                norm_layer(self.key_channels), nn.ReLU(inplace=True)
            ),
        )
        self.f_object = nn.Sequential(
            nn.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.key_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.Sequential(
                norm_layer(self.key_channels), nn.ReLU(inplace=True)
            ),
            nn.Conv2d(
                in_channels=self.key_channels,
                out_channels=self.key_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.Sequential(
                norm_layer(self.key_channels), nn.ReLU(inplace=True)
            ),
        )
        self.f_down = nn.Sequential(
            nn.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.key_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.Sequential(
                norm_layer(self.key_channels), nn.ReLU(inplace=True)
            ),
        )
        self.f_up = nn.Sequential(
            nn.Conv2d(
                in_channels=self.key_channels,
                out_channels=self.in_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.Sequential(norm_layer(self.in_channels), nn.ReLU(inplace=True)),
        )

    def forward(self, x, proxy):
        batch_size, h, w = x.size(0), x.size(2), x.size(3)
        if self.scale > 1:
            x = self.pool(x)

        query = self.f_pixel(x).view(batch_size, self.key_channels, -1)
        query = query.permute(0, 2, 1)
        key = self.f_object(proxy).view(batch_size, self.key_channels, -1)
        value = self.f_down(proxy).view(batch_size, self.key_channels, -1)
        value = value.permute(0, 2, 1)

        sim_map = torch.matmul(query, key)
        sim_map = (self.key_channels**-0.5) * sim_map
        sim_map = F.softmax(sim_map, dim=-1)

        # add bg context ...
        context = torch.matmul(sim_map, value)
        context = context.permute(0, 2, 1).contiguous()
        context = context.view(batch_size, self.key_channels, *x.size()[2:])
        context = self.f_up(context)
        if self.scale > 1:
            context = F.interpolate(
                input=context,
                size=(h, w),
                mode="bilinear",
                align_corners=self.align_corners,
            )

        return context



import torch.nn as nn

from ritm_annotation.model.modifiers import LRMult
from ritm_annotation.utils.serialization import serialize

from .is_model import ISModel
from .modeling.basic_blocks import SepConvHead
from .modeling.deeplab_v3 import DeepLabV3Plus


class DeeplabModel(ISModel):
    @serialize
    def __init__(
        self,
        backbone="resnet50",
        deeplab_ch=256,
        aspp_dropout=0.5,
        backbone_norm_layer=None,
        backbone_lr_mult=0.1,
        norm_layer=nn.BatchNorm2d,
        **kwargs
    ):
        super().__init__(norm_layer=norm_layer, **kwargs)

        self.feature_extractor = DeepLabV3Plus(
            backbone=backbone,
            ch=deeplab_ch,
            project_dropout=aspp_dropout,
            norm_layer=norm_layer,
            backbone_norm_layer=backbone_norm_layer,
        )
        self.feature_extractor.backbone.apply(LRMult(backbone_lr_mult))
        self.head = SepConvHead(
            1,
            in_channels=deeplab_ch,
            mid_channels=deeplab_ch // 2,
            num_layers=2,
            norm_layer=norm_layer,
        )

    def backbone_forward(self, image, coord_features=None):
        backbone_features = self.feature_extractor(image, coord_features)

        return {"instances": self.head(backbone_features[0])}



# flake8: noqa

from ritm_annotation.utils.exp_imports.default import *

MODEL_NAME = "resnet34"


def init_model(cfg, dry_run=False):
    model_cfg = edict()
    model_cfg.crop_size = (320, 480)
    model_cfg.num_max_points = 24
    model_cfg.default_num_epochs = 160

    model = DeeplabModel(
        backbone="resnet34",
        deeplab_ch=128,
        aspp_dropout=0.20,
        use_leaky_relu=True,
        use_rgb_conv=False,
        use_disks=True,
        norm_radius=5,
    )

    model.to(cfg.device)
    model.apply(initializer.XavierGluon(rnd_type="gaussian", magnitude=2.0))
    if not dry_run:
        model.feature_extractor.load_pretrained_weights()

    return model, model_cfg


def get_trainer(model, cfg, model_cfg, dry_run=False, no_dataset=False):
    cfg.batch_size = 28 if cfg.batch_size < 1 else cfg.batch_size
    cfg.val_batch_size = cfg.batch_size
    crop_size = model_cfg.crop_size

    loss_cfg = edict()
    loss_cfg.instance_loss = NormalizedFocalLossSigmoid(alpha=0.5, gamma=2)
    loss_cfg.instance_loss_weight = 1.0

    train_augmentator = Compose(
        [
            UniformRandomResize(scale_range=(0.75, 1.40)),
            HorizontalFlip(),
            PadIfNeeded(
                min_height=crop_size[0], min_width=crop_size[1], border_mode=0
            ),
            RandomCrop(*crop_size),
            RandomBrightnessContrast(
                brightness_limit=(-0.25, 0.25),
                contrast_limit=(-0.15, 0.4),
                p=0.75,
            ),
            RGBShift(
                r_shift_limit=10, g_shift_limit=10, b_shift_limit=10, p=0.75
            ),
        ],
        p=1.0,
    )

    val_augmentator = Compose(
        [
            PadIfNeeded(
                min_height=crop_size[0], min_width=crop_size[1], border_mode=0
            ),
            RandomCrop(*crop_size),
        ],
        p=1.0,
    )

    points_sampler = MultiPointSampler(
        model_cfg.num_max_points,
        prob_gamma=0.8,
        merge_objects_prob=0.15,
        max_num_merged_objects=2,
    )

    if no_dataset:
        trainset = None
        valset = None
    else:
        trainset = LvisDataset(
            cfg.LVIS_PATH,
            split="train",
            augmentator=train_augmentator,
            min_object_area=1000,
            keep_background_prob=0.05,
            points_sampler=points_sampler,
            epoch_len=30000,
            dry_run=dry_run,
        )

        valset = LvisDataset(
            cfg.LVIS_PATH,
            split="val",
            augmentator=val_augmentator,
            min_object_area=1000,
            points_sampler=points_sampler,
            epoch_len=2000,
            dry_run=dry_run,
        )

    optimizer_params = {"lr": 5e-4, "betas": (0.9, 0.999), "eps": 1e-8}

    lr_scheduler = partial(
        torch.optim.lr_scheduler.MultiStepLR, milestones=[140, 155], gamma=0.1
    )
    return ISTrainer(
        model,
        cfg,
        model_cfg,
        loss_cfg,
        trainset,
        valset,
        optimizer="adam",
        optimizer_params=optimizer_params,
        lr_scheduler=lr_scheduler,
        checkpoint_interval=5,
        image_dump_interval=2000,
        metrics=[AdaptiveIoU()],
        max_interactive_points=model_cfg.num_max_points,
        dry_run=dry_run,
    )



from typing import List

import torch

from ritm_annotation.inference.clicker import Click

from .base import BaseTransform


class AddHorizontalFlip(BaseTransform):
    def transform(self, image_nd, clicks_lists: List[List[Click]]):
        assert len(image_nd.shape) == 4
        image_nd = torch.cat([image_nd, torch.flip(image_nd, dims=[3])], dim=0)

        image_width = image_nd.shape[3]
        clicks_lists_flipped = []
        for clicks_list in clicks_lists:
            clicks_list_flipped = [
                click.copy(
                    coords=(click.coords[0], image_width - click.coords[1] - 1)
                )
                for click in clicks_list
            ]
            clicks_lists_flipped.append(clicks_list_flipped)
        clicks_lists = clicks_lists + clicks_lists_flipped

        return image_nd, clicks_lists

    def inv_transform(self, prob_map):
        assert len(prob_map.shape) == 4 and prob_map.shape[0] % 2 == 0
        num_maps = prob_map.shape[0] // 2
        prob_map, prob_map_flipped = prob_map[:num_maps], prob_map[num_maps:]

        return 0.5 * (prob_map + torch.flip(prob_map_flipped, dims=[3]))

    def get_state(self):
        return None

    def set_state(self, state):
        pass

    def reset(self):
        pass



import numpy as np
import torch
import torch.nn.functional as F
from scipy.optimize import fmin_l_bfgs_b

from .base import BasePredictor


class BRSBasePredictor(BasePredictor):
    def __init__(
        self, model, device, opt_functor, optimize_after_n_clicks=1, **kwargs
    ):
        super().__init__(model, device, **kwargs)
        self.optimize_after_n_clicks = optimize_after_n_clicks
        self.opt_functor = opt_functor

        self.opt_data = None
        self.input_data = None

    def set_input_image(self, image):
        super().set_input_image(image)
        self.opt_data = None
        self.input_data = None

    def _get_clicks_maps_nd(self, clicks_lists, image_shape, radius=1):
        pos_clicks_map = np.zeros(
            (len(clicks_lists), 1) + image_shape, dtype=np.float32
        )
        neg_clicks_map = np.zeros(
            (len(clicks_lists), 1) + image_shape, dtype=np.float32
        )

        for list_indx, clicks_list in enumerate(clicks_lists):
            for click in clicks_list:
                y, x = click.coords
                y, x = int(round(y)), int(round(x))
                y1, x1 = y - radius, x - radius
                y2, x2 = y + radius + 1, x + radius + 1

                if click.is_positive:
                    pos_clicks_map[list_indx, 0, y1:y2, x1:x2] = True
                else:
                    neg_clicks_map[list_indx, 0, y1:y2, x1:x2] = True

        with torch.no_grad():
            pos_clicks_map = torch.from_numpy(pos_clicks_map).to(self.device)
            neg_clicks_map = torch.from_numpy(neg_clicks_map).to(self.device)

        return pos_clicks_map, neg_clicks_map

    def get_states(self):
        return {
            "transform_states": self._get_transform_states(),
            "opt_data": self.opt_data,
        }

    def set_states(self, states):
        self._set_transform_states(states["transform_states"])
        self.opt_data = states["opt_data"]


class FeatureBRSPredictor(BRSBasePredictor):
    def __init__(
        self,
        model,
        device,
        opt_functor,
        insertion_mode="after_deeplab",
        **kwargs
    ):
        super().__init__(model, device, opt_functor=opt_functor, **kwargs)
        self.insertion_mode = insertion_mode
        self._c1_features = None

        if self.insertion_mode == "after_deeplab":
            self.num_channels = model.feature_extractor.ch
        elif self.insertion_mode == "after_c4":
            self.num_channels = model.feature_extractor.aspp_in_channels
        elif self.insertion_mode == "after_aspp":
            self.num_channels = model.feature_extractor.ch + 32
        else:
            raise NotImplementedError

    def _get_prediction(self, image_nd, clicks_lists, is_image_changed):
        points_nd = self.get_points_nd(clicks_lists)
        pos_mask, neg_mask = self._get_clicks_maps_nd(
            clicks_lists, image_nd.shape[2:]
        )

        num_clicks = len(clicks_lists[0])
        bs = image_nd.shape[0] // 2 if self.with_flip else image_nd.shape[0]

        if (
            self.opt_data is None
            or self.opt_data.shape[0] // (2 * self.num_channels) != bs
        ):
            self.opt_data = np.zeros(
                (bs * 2 * self.num_channels), dtype=np.float32
            )

        if (
            num_clicks <= self.net_clicks_limit
            or is_image_changed
            or self.input_data is None
        ):
            self.input_data = self._get_head_input(image_nd, points_nd)

        def get_prediction_logits(scale, bias):
            scale = scale.view(bs, -1, 1, 1)
            bias = bias.view(bs, -1, 1, 1)
            if self.with_flip:
                scale = scale.repeat(2, 1, 1, 1)
                bias = bias.repeat(2, 1, 1, 1)

            scaled_backbone_features = self.input_data * scale
            scaled_backbone_features = scaled_backbone_features + bias
            if self.insertion_mode == "after_c4":
                x = self.net.feature_extractor.aspp(scaled_backbone_features)
                x = F.interpolate(
                    x,
                    mode="bilinear",
                    size=self._c1_features.size()[2:],
                    align_corners=True,
                )
                x = torch.cat((x, self._c1_features), dim=1)
                scaled_backbone_features = self.net.feature_extractor.head(x)
            elif self.insertion_mode == "after_aspp":
                scaled_backbone_features = self.net.feature_extractor.head(
                    scaled_backbone_features
                )

            pred_logits = self.net.head(scaled_backbone_features)
            pred_logits = F.interpolate(
                pred_logits,
                size=image_nd.size()[2:],
                mode="bilinear",
                align_corners=True,
            )
            return pred_logits

        self.opt_functor.init_click(
            get_prediction_logits, pos_mask, neg_mask, self.device
        )
        if num_clicks > self.optimize_after_n_clicks:
            opt_result = fmin_l_bfgs_b(
                func=self.opt_functor,
                x0=self.opt_data,
                **self.opt_functor.optimizer_params
            )
            self.opt_data = opt_result[0]

        with torch.no_grad():
            if self.opt_functor.best_prediction is not None:
                opt_pred_logits = self.opt_functor.best_prediction
            else:
                opt_data_nd = torch.from_numpy(self.opt_data).to(self.device)
                opt_vars, _ = self.opt_functor.unpack_opt_params(opt_data_nd)
                opt_pred_logits = get_prediction_logits(*opt_vars)

        return opt_pred_logits

    def _get_head_input(self, image_nd, points):
        with torch.no_grad():
            image_nd, prev_mask = self.net.prepare_input(image_nd)
            coord_features = self.net.get_coord_features(
                image_nd, prev_mask, points
            )

            if self.net.rgb_conv is not None:
                x = self.net.rgb_conv(
                    torch.cat((image_nd, coord_features), dim=1)
                )
                additional_features = None
            elif hasattr(self.net, "maps_transform"):
                x = image_nd
                additional_features = self.net.maps_transform(coord_features)

            if (
                self.insertion_mode == "after_c4"
                or self.insertion_mode == "after_aspp"
            ):
                c1, _, c3, c4 = self.net.feature_extractor.backbone(
                    x, additional_features
                )
                c1 = self.net.feature_extractor.skip_project(c1)

                if self.insertion_mode == "after_aspp":
                    x = self.net.feature_extractor.aspp(c4)
                    x = F.interpolate(
                        x,
                        size=c1.size()[2:],
                        mode="bilinear",
                        align_corners=True,
                    )
                    x = torch.cat((x, c1), dim=1)
                    backbone_features = x
                else:
                    backbone_features = c4
                    self._c1_features = c1
            else:
                backbone_features = self.net.feature_extractor(
                    x, additional_features
                )[0]

        return backbone_features


class HRNetFeatureBRSPredictor(BRSBasePredictor):
    def __init__(
        self, model, device, opt_functor, insertion_mode="A", **kwargs
    ):
        super().__init__(model, device, opt_functor=opt_functor, **kwargs)
        self.insertion_mode = insertion_mode
        self._c1_features = None

        if self.insertion_mode == "A":
            self.num_channels = sum(
                k * model.feature_extractor.width for k in [1, 2, 4, 8]
            )
        elif self.insertion_mode == "C":
            self.num_channels = 2 * model.feature_extractor.ocr_width
        else:
            raise NotImplementedError

    def _get_prediction(self, image_nd, clicks_lists, is_image_changed):
        points_nd = self.get_points_nd(clicks_lists)
        pos_mask, neg_mask = self._get_clicks_maps_nd(
            clicks_lists, image_nd.shape[2:]
        )
        num_clicks = len(clicks_lists[0])
        bs = image_nd.shape[0] // 2 if self.with_flip else image_nd.shape[0]

        if (
            self.opt_data is None
            or self.opt_data.shape[0] // (2 * self.num_channels) != bs
        ):
            self.opt_data = np.zeros(
                (bs * 2 * self.num_channels), dtype=np.float32
            )

        if (
            num_clicks <= self.net_clicks_limit
            or is_image_changed
            or self.input_data is None
        ):
            self.input_data = self._get_head_input(image_nd, points_nd)

        def get_prediction_logits(scale, bias):
            scale = scale.view(bs, -1, 1, 1)
            bias = bias.view(bs, -1, 1, 1)
            if self.with_flip:
                scale = scale.repeat(2, 1, 1, 1)
                bias = bias.repeat(2, 1, 1, 1)

            scaled_backbone_features = self.input_data * scale
            scaled_backbone_features = scaled_backbone_features + bias
            if self.insertion_mode == "A":
                if self.net.feature_extractor.ocr_width > 0:
                    out_aux = self.net.feature_extractor.aux_head(
                        scaled_backbone_features
                    )
                    feats = self.net.feature_extractor.conv3x3_ocr(
                        scaled_backbone_features
                    )

                    context = self.net.feature_extractor.ocr_gather_head(
                        feats, out_aux
                    )
                    feats = self.net.feature_extractor.ocr_distri_head(
                        feats, context
                    )
                else:
                    feats = scaled_backbone_features
                pred_logits = self.net.feature_extractor.cls_head(feats)
            elif self.insertion_mode == "C":
                pred_logits = self.net.feature_extractor.cls_head(
                    scaled_backbone_features
                )
            else:
                raise NotImplementedError

            pred_logits = F.interpolate(
                pred_logits,
                size=image_nd.size()[2:],
                mode="bilinear",
                align_corners=True,
            )
            return pred_logits

        self.opt_functor.init_click(
            get_prediction_logits, pos_mask, neg_mask, self.device
        )
        if num_clicks > self.optimize_after_n_clicks:
            opt_result = fmin_l_bfgs_b(
                func=self.opt_functor,
                x0=self.opt_data,
                **self.opt_functor.optimizer_params
            )
            self.opt_data = opt_result[0]

        with torch.no_grad():
            if self.opt_functor.best_prediction is not None:
                opt_pred_logits = self.opt_functor.best_prediction
            else:
                opt_data_nd = torch.from_numpy(self.opt_data).to(self.device)
                opt_vars, _ = self.opt_functor.unpack_opt_params(opt_data_nd)
                opt_pred_logits = get_prediction_logits(*opt_vars)

        return opt_pred_logits

    def _get_head_input(self, image_nd, points):
        with torch.no_grad():
            image_nd, prev_mask = self.net.prepare_input(image_nd)
            coord_features = self.net.get_coord_features(
                image_nd, prev_mask, points
            )

            if self.net.rgb_conv is not None:
                x = self.net.rgb_conv(
                    torch.cat((image_nd, coord_features), dim=1)
                )
                additional_features = None
            elif hasattr(self.net, "maps_transform"):
                x = image_nd
                additional_features = self.net.maps_transform(coord_features)

            feats = self.net.feature_extractor.compute_hrnet_feats(
                x, additional_features
            )

            if self.insertion_mode == "A":
                backbone_features = feats
            elif self.insertion_mode == "C":
                out_aux = self.net.feature_extractor.aux_head(feats)
                feats = self.net.feature_extractor.conv3x3_ocr(feats)

                context = self.net.feature_extractor.ocr_gather_head(
                    feats, out_aux
                )
                backbone_features = self.net.feature_extractor.ocr_distri_head(
                    feats, context
                )
            else:
                raise NotImplementedError

        return backbone_features


class InputBRSPredictor(BRSBasePredictor):
    def __init__(
        self, model, device, opt_functor, optimize_target="rgb", **kwargs
    ):
        super().__init__(model, device, opt_functor=opt_functor, **kwargs)
        self.optimize_target = optimize_target

    def _get_prediction(self, image_nd, clicks_lists, is_image_changed):
        points_nd = self.get_points_nd(clicks_lists)
        pos_mask, neg_mask = self._get_clicks_maps_nd(
            clicks_lists, image_nd.shape[2:]
        )
        num_clicks = len(clicks_lists[0])

        if self.opt_data is None or is_image_changed:
            if self.optimize_target == "dmaps":
                opt_channels = (
                    self.net.coord_feature_ch - 1
                    if self.net.with_prev_mask
                    else self.net.coord_feature_ch
                )
            else:
                opt_channels = 3
            bs = (
                image_nd.shape[0] // 2 if self.with_flip else image_nd.shape[0]
            )
            self.opt_data = torch.zeros(
                (bs, opt_channels, image_nd.shape[2], image_nd.shape[3]),
                device=self.device,
                dtype=torch.float32,
            )

        def get_prediction_logits(opt_bias):
            input_image, prev_mask = self.net.prepare_input(image_nd)
            dmaps = self.net.get_coord_features(
                input_image, prev_mask, points_nd
            )

            if self.optimize_target == "rgb":
                input_image = input_image + opt_bias
            elif self.optimize_target == "dmaps":
                if self.net.with_prev_mask:
                    dmaps[:, 1:, :, :] = dmaps[:, 1:, :, :] + opt_bias
                else:
                    dmaps = dmaps + opt_bias

            if self.net.rgb_conv is not None:
                x = self.net.rgb_conv(torch.cat((input_image, dmaps), dim=1))
                if self.optimize_target == "all":
                    x = x + opt_bias
                coord_features = None
            elif hasattr(self.net, "maps_transform"):
                x = input_image
                coord_features = self.net.maps_transform(dmaps)

            pred_logits = self.net.backbone_forward(
                x, coord_features=coord_features
            )["instances"]
            pred_logits = F.interpolate(
                pred_logits,
                size=image_nd.size()[2:],
                mode="bilinear",
                align_corners=True,
            )

            return pred_logits

        self.opt_functor.init_click(
            get_prediction_logits,
            pos_mask,
            neg_mask,
            self.device,
            shape=self.opt_data.shape,
        )
        if num_clicks > self.optimize_after_n_clicks:
            opt_result = fmin_l_bfgs_b(
                func=self.opt_functor,
                x0=self.opt_data.cpu().numpy().ravel(),
                **self.opt_functor.optimizer_params
            )

            self.opt_data = (
                torch.from_numpy(opt_result[0])
                .view(self.opt_data.shape)
                .to(self.device)
            )

        with torch.no_grad():
            if self.opt_functor.best_prediction is not None:
                opt_pred_logits = self.opt_functor.best_prediction
            else:
                opt_vars, _ = self.opt_functor.unpack_opt_params(self.opt_data)
                opt_pred_logits = get_prediction_logits(*opt_vars)

        return opt_pred_logits



import torch
from torch import distributed as dist
from torch.utils import data


def get_rank():
    if not dist.is_available() or not dist.is_initialized():
        return 0
    return dist.get_rank()


def synchronize():
    if (
        not dist.is_available()
        or not dist.is_initialized()
        or dist.get_world_size() == 1
    ):
        return
    dist.barrier()


def get_world_size():
    if not dist.is_available() or not dist.is_initialized():
        return 1

    return dist.get_world_size()


def reduce_loss_dict(loss_dict):
    world_size = get_world_size()

    if world_size < 2:
        return loss_dict

    with torch.no_grad():
        keys = []
        losses = []

        for k in loss_dict.keys():
            keys.append(k)
            losses.append(loss_dict[k])

        losses = torch.stack(losses, 0)
        dist.reduce(losses, dst=0)

        if dist.get_rank() == 0:
            losses /= world_size

        reduced_losses = {k: v for k, v in zip(keys, losses)}

    return reduced_losses


def get_sampler(dataset, shuffle, distributed):
    if distributed:
        return data.distributed.DistributedSampler(dataset, shuffle=shuffle)

    if shuffle:
        return data.RandomSampler(dataset)
    else:
        return data.SequentialSampler(dataset)


def get_dp_wrapper(distributed):
    class DPWrapper(
        torch.nn.parallel.DistributedDataParallel
        if distributed
        else torch.nn.DataParallel
    ):
        def __getattr__(self, name):
            try:
                return super().__getattr__(name)
            except AttributeError:
                return getattr(self.module, name)

    return DPWrapper



# flake8: noqa E501

import logging
import tkinter as tk
from pathlib import Path
from sys import exit
from tkinter import messagebox, ttk
import json

import cv2
import numpy as np
import torch
from easydict import EasyDict as edict
from PIL import Image

from ritm_annotation.cli.annotate.canvas import CanvasImage
from ritm_annotation.cli.annotate.controller import InteractiveController
from ritm_annotation.cli.annotate.wrappers import (
    BoundedNumericalEntry,
    FocusButton,
    FocusCheckButton,
    FocusHorizontalScale,
    FocusLabelFrame,
)
from ritm_annotation.inference.utils import find_checkpoint, load_is_model
from ritm_annotation.utils.misc import ignore_params_then_call, get_default_weight

logger = logging.getLogger(__name__)


class InteractiveDemoApp(ttk.Frame):
    def __init__(self, master, args, model, tasks_iterator):
        logger.info("Initializing...")
        super().__init__(master)
        self.tasks = list(tasks_iterator)
        self._current_task_idx = 0
        self.master = master
        master.title(
            _(
                "Reviving Iterative Training with Mask Guidance for Interactive Segmentation"
            )  # noqa: E501
        )
        master.withdraw()
        master.update_idletasks()
        x = (master.winfo_screenwidth() - master.winfo_reqwidth()) / 2
        y = (master.winfo_screenheight() - master.winfo_reqheight()) / 2
        master.geometry("+%d+%d" % (x, y))
        self.pack(fill="both", expand=True)

        self.brs_modes = [
            "NoBRS",
            "RGB-BRS",
            "DistMap-BRS",
            "f-BRS-A",
            "f-BRS-B",
            "f-BRS-C",
        ]
        self.limit_longest_size = args.limit_longest_size

        self.controller = InteractiveController(
            model,
            args.device,
            predictor_params={"brs_mode": "NoBRS"},
            update_image_callback=self._update_image,
        )

        self._init_state()
        self._add_menu()
        self._add_canvas()
        self._add_buttons()

        master.bind("<Escape>", lambda event: exit(0))
        master.bind(
            "j",
            lambda event: self.state["visualize_mask"].set(
                not self.state["visualize_mask"].get()
            ),
        )
        master.bind(
            "k",
            lambda event: self.state["visualize_blender_contrast"].set(
                not self.state["visualize_blender_contrast"].get()
            ),
        )
        master.bind("l", lambda event: self._save_mask_callback())
        master.bind("h", lambda event: self._undo_click())
        master.bind("m", lambda event: self._reset_last_object())
        master.bind("n", lambda event: self._load_image_callback())
        master.bind("b", lambda event: self._goto_previous_task())

        self.state["zoomin_params"]["skip_clicks"].trace(
            mode="w", callback=self._reset_predictor
        )
        self.state["zoomin_params"]["target_size"].trace(
            mode="w", callback=self._reset_predictor
        )
        self.state["zoomin_params"]["expansion_ratio"].trace(
            mode="w", callback=self._reset_predictor
        )
        self.state["predictor_params"]["net_clicks_limit"].trace(
            mode="w", callback=self._change_brs_mode
        )
        self.state["lbfgs_max_iters"].trace(
            mode="w", callback=self._change_brs_mode
        )
        self._change_brs_mode()
        self._current_task = None

    def _handle_classe_finalizada(self, classe):
        messagebox.showwarning(
            _("Class finished"),
            _("The class '{class_name}' was finalized").format(
                class_name=classe
            ),
        )

    def goto_task(self, idx=None):
        if idx is None:
            idx = self._current_task_idx + 1
        if idx >= len(self.tasks):
            raise StopIteration()
        print('goto_task', idx, self.tasks[idx])
        self._current_task = self.tasks[idx]
        self._current_task_idx = idx
        return self._current_task

    def _goto_previous_task(self):
        if self._current_task_idx < 1:
            return
        self._current_task_idx -= 1
        self._load_image_callback(current_task=self.goto_task(self._current_task_idx))

    def _get_current_task(self, ask_next=False):
        current_class = (
            self._current_task.class_name
            if self._current_task is not None
            else _("* no one *")
        )
        try:
            if self._current_task is None or ask_next:
                logger.debug(_("Pulling next task from the iterator"))
                next_task = self.goto_task()
                if current_class != next_task.class_name:
                    self._handle_classe_finalizada(current_class)
                self._current_task = next_task
            self.task_label.config(
                text=f"class:{self._current_task.class_name} {self._current_task.image}"  # noqa: E501
            )
            return self._current_task
        except StopIteration:
            self._handle_classe_finalizada(current_class)
            exit(0)

    def _init_state(self):
        self.state = {
            "zoomin_params": {
                "use_zoom_in": tk.BooleanVar(value=True),
                "fixed_crop": tk.BooleanVar(value=True),
                "skip_clicks": tk.IntVar(value=-1),
                "target_size": tk.IntVar(
                    value=min(400, self.limit_longest_size)
                ),
                "expansion_ratio": tk.DoubleVar(value=1.4),
            },
            "predictor_params": {"net_clicks_limit": tk.IntVar(value=8)},
            "brs_mode": tk.StringVar(value="NoBRS"),
            "prob_thresh": tk.DoubleVar(value=0.5),
            "lbfgs_max_iters": tk.IntVar(value=20),
            "alpha_blend": tk.DoubleVar(value=0.5),
            "click_radius": tk.IntVar(value=3),
            "visualize_mask": tk.BooleanVar(value=False),
            "visualize_blender_contrast": tk.BooleanVar(value=False),
        }

    def _add_menu(self):
        self.menubar = FocusLabelFrame(self, bd=1)
        self.menubar.pack(side=tk.TOP, fill="x")

        button = FocusButton(
            self.menubar,
            text=_("Load image") +  " (n)",
            command=self._load_image_callback,
        )
        button.pack(side=tk.LEFT)
        self.save_mask_btn = FocusButton(
            self.menubar,
            text=_("Save mask") +  " (l)",
            command=self._save_mask_callback,
        )
        self.save_mask_btn.pack(side=tk.LEFT)
        self.save_mask_btn.configure(state=tk.DISABLED)

        self.previous_image_btn = FocusButton(
            self.menubar, text=_("Previous image") + " (b)", command=self._goto_previous_task
        )
        self.previous_image_btn.pack(side=tk.LEFT)
        # self.previous_image_btn.configure(state=tk.DISABLED)

        self.load_mask_btn = FocusButton(
            self.menubar, text=_("Load mask"), command=self._load_mask_callback
        )
        self.load_mask_btn.pack(side=tk.LEFT)
        self.load_mask_btn.configure(state=tk.DISABLED)

        button = FocusButton(
            self.menubar, text=_("About"), command=self._about_callback
        )
        button.pack(side=tk.LEFT)
        button = FocusButton(
            self.menubar, text=_("Exit"), command=self.master.quit
        )
        button.pack(side=tk.LEFT)

        self.task_label = tk.Label(self.menubar, text=_("* status *"))
        self.task_label.pack(side=tk.LEFT)

    def _add_canvas(self):
        self.canvas_frame = FocusLabelFrame(self, text="Image")
        self.canvas_frame.rowconfigure(0, weight=1)
        self.canvas_frame.columnconfigure(0, weight=1)

        self.canvas = tk.Canvas(
            self.canvas_frame,
            highlightthickness=0,
            cursor="hand1",
            width=400,
            height=400,
        )
        self.canvas.grid(row=0, column=0, sticky="nswe", padx=5, pady=5)

        self.image_on_canvas = None
        self.canvas_frame.pack(
            side=tk.LEFT, fill="both", expand=True, padx=5, pady=5
        )

    def _add_buttons(self):
        self.control_frame = FocusLabelFrame(self, text=_("Controls"))
        self.control_frame.pack(side=tk.TOP, fill="x", padx=5, pady=5)
        master = self.control_frame

        FocusCheckButton(
            master,
            text=f"{_('Mostrar máscara ao invés da imagem')} (j)",
            variable=self.state["visualize_mask"],
        ).pack(side=tk.TOP)
        self.state["visualize_mask"].trace(
            mode="w", callback=ignore_params_then_call(self._update_image)
        )
        FocusCheckButton(
            master,
            text=f"{_('Mostrar sobreposição da máscara com mais contraste')} (k)",
            variable=self.state["visualize_blender_contrast"],
        ).pack(side=tk.TOP)
        self.state["visualize_blender_contrast"].trace(
            mode="w", callback=ignore_params_then_call(self._update_image)
        )

        self.clicks_options_frame = FocusLabelFrame(
            master, text=_("Clicks management")
        )
        self.clicks_options_frame.pack(side=tk.TOP, fill=tk.X, padx=10, pady=3)
        self.finish_object_button = FocusButton(
            self.clicks_options_frame,
            text=_("Finish\nobject"),
            bg="#b6d7a8",
            fg="black",
            width=10,
            height=2,
            state=tk.DISABLED,
            command=self.controller.finish_object,
        )
        self.finish_object_button.pack(
            side=tk.LEFT, fill=tk.X, padx=10, pady=3
        )
        self.undo_click_button = FocusButton(
            self.clicks_options_frame,
            text=f"{_('Undo click')} (h)",
            bg="#ffe599",
            fg="black",
            width=10,
            height=2,
            state=tk.DISABLED,
            command=self._undo_click,
        )
        self.undo_click_button.pack(side=tk.LEFT, fill=tk.X, padx=10, pady=3)
        self.reset_clicks_button = FocusButton(
            self.clicks_options_frame,
            text=f"{_('Reset clicks')} (m)",
            bg="#ea9999",
            fg="black",
            width=10,
            height=2,
            state=tk.DISABLED,
            command=self._reset_last_object,
        )
        self.reset_clicks_button.pack(side=tk.LEFT, fill=tk.X, padx=10, pady=3)

        self.zoomin_options_frame = FocusLabelFrame(
            master, text=_("ZoomIn options")
        )
        self.zoomin_options_frame.pack(side=tk.TOP, fill=tk.X, padx=10, pady=3)
        FocusCheckButton(
            self.zoomin_options_frame,
            text=_("Use ZoomIn"),
            command=self._reset_predictor,
            variable=self.state["zoomin_params"]["use_zoom_in"],
        ).grid(row=0, column=0, padx=10)
        FocusCheckButton(
            self.zoomin_options_frame,
            text=_("Fixed crop"),
            command=self._reset_predictor,
            variable=self.state["zoomin_params"]["fixed_crop"],
        ).grid(row=1, column=0, padx=10)
        tk.Label(self.zoomin_options_frame, text=_("Skip clicks")).grid(
            row=0, column=1, pady=1, sticky="e"
        )
        tk.Label(self.zoomin_options_frame, text=_("Target size")).grid(
            row=1, column=1, pady=1, sticky="e"
        )
        tk.Label(self.zoomin_options_frame, text=_("Expand ratio")).grid(
            row=2, column=1, pady=1, sticky="e"
        )
        BoundedNumericalEntry(
            self.zoomin_options_frame,
            variable=self.state["zoomin_params"]["skip_clicks"],
            min_value=-1,
            max_value=None,
            vartype=int,
            name="zoom_in_skip_clicks",
        ).grid(row=0, column=2, padx=10, pady=1, sticky="w")
        BoundedNumericalEntry(
            self.zoomin_options_frame,
            variable=self.state["zoomin_params"]["target_size"],
            min_value=100,
            max_value=self.limit_longest_size,
            vartype=int,
            name="zoom_in_target_size",
        ).grid(row=1, column=2, padx=10, pady=1, sticky="w")
        BoundedNumericalEntry(
            self.zoomin_options_frame,
            variable=self.state["zoomin_params"]["expansion_ratio"],
            min_value=1.0,
            max_value=2.0,
            vartype=float,
            name="zoom_in_expansion_ratio",
        ).grid(row=2, column=2, padx=10, pady=1, sticky="w")
        self.zoomin_options_frame.columnconfigure((0, 1, 2), weight=1)

        self.brs_options_frame = FocusLabelFrame(master, text=_("BRS options"))
        self.brs_options_frame.pack(side=tk.TOP, fill=tk.X, padx=10, pady=3)
        menu = tk.OptionMenu(
            self.brs_options_frame,
            self.state["brs_mode"],
            *self.brs_modes,
            command=self._change_brs_mode,
        )
        menu.config(width=11)
        menu.grid(rowspan=2, column=0, padx=10)
        self.net_clicks_label = tk.Label(
            self.brs_options_frame, text=_("Network clicks")
        )
        self.net_clicks_label.grid(row=0, column=1, pady=2, sticky="e")
        self.net_clicks_entry = BoundedNumericalEntry(
            self.brs_options_frame,
            variable=self.state["predictor_params"]["net_clicks_limit"],
            min_value=0,
            max_value=None,
            vartype=int,
            allow_inf=True,
            name="net_clicks_limit",
        )
        self.net_clicks_entry.grid(
            row=0, column=2, padx=10, pady=2, sticky="w"
        )
        self.lbfgs_iters_label = tk.Label(
            self.brs_options_frame, text=_("L-BFGS\nmax iterations")
        )
        self.lbfgs_iters_label.grid(row=1, column=1, pady=2, sticky="e")
        self.lbfgs_iters_entry = BoundedNumericalEntry(
            self.brs_options_frame,
            variable=self.state["lbfgs_max_iters"],
            min_value=1,
            max_value=1000,
            vartype=int,
            name="lbfgs_max_iters",
        )
        self.lbfgs_iters_entry.grid(
            row=1, column=2, padx=10, pady=2, sticky="w"
        )
        self.brs_options_frame.columnconfigure((0, 1), weight=1)

        self.prob_thresh_frame = FocusLabelFrame(
            master, text=_("Predictions threshold")
        )
        self.prob_thresh_frame.pack(side=tk.TOP, fill=tk.X, padx=10, pady=3)
        FocusHorizontalScale(
            self.prob_thresh_frame,
            from_=0.0,
            to=1.0,
            command=self._update_prob_thresh,
            variable=self.state["prob_thresh"],
        ).pack(padx=10)

        self.alpha_blend_frame = FocusLabelFrame(
            master, text=_("Alpha blending coefficient")
        )
        self.alpha_blend_frame.pack(side=tk.TOP, fill=tk.X, padx=10, pady=3)
        FocusHorizontalScale(
            self.alpha_blend_frame,
            from_=0.0,
            to=1.0,
            command=self._update_blend_alpha,
            variable=self.state["alpha_blend"],
        ).pack(padx=10, anchor=tk.CENTER)

        self.click_radius_frame = FocusLabelFrame(
            master, text=_("Visualisation click radius")
        )
        self.click_radius_frame.pack(side=tk.TOP, fill=tk.X, padx=10, pady=3)
        FocusHorizontalScale(
            self.click_radius_frame,
            from_=0,
            to=7,
            resolution=1,
            command=self._update_click_radius,
            variable=self.state["click_radius"],
        ).pack(padx=10, anchor=tk.CENTER)

    def _load_image_callback(self, current_task=None):
        self.menubar.focus_set()
        if self._check_entry(self):
            if current_task is None:
                current_task = self._get_current_task(ask_next=True)
            filename = str(current_task.image.resolve())
            logger.info(_("Loading '{filename}'").format(filename=filename))
            image = cv2.cvtColor(cv2.imread(filename), cv2.COLOR_BGR2RGB)
            self.controller.set_image(image)
            points_json_name = current_task.class_name + ".json"
            points_json = current_task.output_dir / points_json_name
            if current_task.seed_mask is not None:
                points_json = current_task.seed_mask.parent / points_json_name
                if not points_json.exists():
                    points_json = current_task.output_dir / points_json_name
            if points_json.exists():
                with points_json.open('r') as f:
                    data = json.load(f)
                    for item in data:
                        self._click_callback(item['is_positive'], item['coords'][1], item['coords'][0])
                    # from ritm_annotation.inference.clicker import Click
                    # data = [Click(item['is_positive'], item['coords'][1], item['coords'][0]) for item in data]
            else:
                if current_task.seed_mask is not None:
                    if current_task.seed_mask.exists():
                        mask = cv2.imread(str(current_task.seed_mask), 0)
                        self.controller.set_mask(mask)
            self.save_mask_btn.configure(state=tk.NORMAL)
            self.load_mask_btn.configure(state=tk.NORMAL)
            self._update_image()
            self._reset_predictor()

    def _undo_click(self):
        self.controller.undo_click()
        self._update_image()

    def _save_mask_callback(self):
        self.menubar.focus_set()
        if self._check_entry(self):
            mask = self.controller.result_mask
            if mask is None:
                return
            current_task = self._get_current_task()
            filename = str(current_task.mask_output.resolve())
            mask = (mask > 0) * 255
            logger.info(_("Saving '{filename}'").format(filename=filename))
            cv2.imwrite(filename, mask)
            points_json = current_task.output_dir / (current_task.class_name + ".json")
            with points_json.open('w') as f:
                # json.dump(self.controller.states, f)
                json.dump([item.to_json() for item in self.controller.clicker.get_state()], f)

    def _load_mask_callback(self):
        if not self.controller.net.with_prev_mask:
            messagebox.showwarning(
                _("Warning"),
                _(
                    "The current model doesn't support loading external masks. Please use ITER-M models for that purpose."
                ),
            )
            return

        self.menubar.focus_set()
        if self._check_entry(self):
            current_task = self._get_current_task()
            filename = str(current_task.mask_output.resolve())
            logger.debug(_("Loading '{filename}'").format(filename=filename))
            mask = cv2.imread(filename)[:, :, 0] > 127
            self.controller.set_mask(mask)
            self._update_image()

    def _about_callback(self):
        self.menubar.focus_set()

        text = [
            _("Developed by:"),
            "K.Sofiiuk and I. Petrov",
            "The MIT License, 2021",
        ]

        messagebox.showinfo(_("About Demo"), "\n".join(text))

    def _reset_last_object(self):
        self.state["alpha_blend"].set(0.5)
        self.state["prob_thresh"].set(0.5)
        self.controller.reset_last_object()

    def _update_prob_thresh(self, value):
        if self.controller.is_incomplete_mask:
            self.controller.prob_thresh = self.state["prob_thresh"].get()
            self._update_image()

    def _update_blend_alpha(self, value):
        self._update_image()

    def _update_click_radius(self, *args):
        if self.image_on_canvas is None:
            return

        self._update_image()

    def _change_brs_mode(self, *args):
        if self.state["brs_mode"].get() == "NoBRS":
            self.net_clicks_entry.set("INF")
            self.net_clicks_entry.configure(state=tk.DISABLED)
            self.net_clicks_label.configure(state=tk.DISABLED)
            self.lbfgs_iters_entry.configure(state=tk.DISABLED)
            self.lbfgs_iters_label.configure(state=tk.DISABLED)
        else:
            if self.net_clicks_entry.get() == "INF":
                self.net_clicks_entry.set(8)
            self.net_clicks_entry.configure(state=tk.NORMAL)
            self.net_clicks_label.configure(state=tk.NORMAL)
            self.lbfgs_iters_entry.configure(state=tk.NORMAL)
            self.lbfgs_iters_label.configure(state=tk.NORMAL)

        self._reset_predictor()

    def _reset_predictor(self, *args, **kwargs):
        brs_mode = self.state["brs_mode"].get()
        prob_thresh = self.state["prob_thresh"].get()
        net_clicks_limit = (
            None
            if brs_mode == "NoBRS"
            else self.state["predictor_params"]["net_clicks_limit"].get()
        )

        if self.state["zoomin_params"]["use_zoom_in"].get():
            zoomin_params = {
                "skip_clicks": self.state["zoomin_params"][
                    "skip_clicks"
                ].get(),
                "target_size": self.state["zoomin_params"][
                    "target_size"
                ].get(),
                "expansion_ratio": self.state["zoomin_params"][
                    "expansion_ratio"
                ].get(),
            }
            if self.state["zoomin_params"]["fixed_crop"].get():
                zoomin_params["target_size"] = (
                    zoomin_params["target_size"],
                    zoomin_params["target_size"],
                )
        else:
            zoomin_params = None

        predictor_params = {
            "brs_mode": brs_mode,
            "prob_thresh": prob_thresh,
            "zoom_in_params": zoomin_params,
            "predictor_params": {
                "net_clicks_limit": net_clicks_limit,
                "max_size": self.limit_longest_size,
            },
            "brs_opt_func_params": {"min_iou_diff": 1e-3},
            "lbfgs_params": {"maxfun": self.state["lbfgs_max_iters"].get()},
        }
        self.controller.reset_predictor(predictor_params)

    def _click_callback(self, is_positive, x, y):
        self.canvas.focus_set()

        if self.image_on_canvas is None:
            messagebox.showwarning(
                _("Warning"), _("Please load an image first")
            )
            return

        if self._check_entry(self):
            self.controller.add_click(x, y, is_positive)

    def _update_image(self, reset_canvas=False):
        image = self.controller.get_visualization(
            alpha_blend=self.state["alpha_blend"].get(),
            click_radius=self.state["click_radius"].get(),
        )
        image_mask = (
            np.array(
                cv2.cvtColor(self.controller.result_mask, cv2.COLOR_GRAY2RGB),
                dtype="uint8",
            )
            * 255
        )
        if self.state["visualize_mask"].get():
            image = image_mask

        if self.state["visualize_blender_contrast"].get():
            image = (image // 2) + (image_mask // 2)

        if self.image_on_canvas is None:
            self.image_on_canvas = CanvasImage(self.canvas_frame, self.canvas)
            self.image_on_canvas.register_click_callback(self._click_callback)

        self._set_click_dependent_widgets_state()
        if image is not None:
            self.image_on_canvas.reload_image(
                Image.fromarray(image), reset_canvas
            )

    def _set_click_dependent_widgets_state(self):
        after_1st_click_state = (
            tk.NORMAL if self.controller.is_incomplete_mask else tk.DISABLED
        )
        before_1st_click_state = (
            tk.DISABLED if self.controller.is_incomplete_mask else tk.NORMAL
        )

        self.finish_object_button.configure(state=after_1st_click_state)
        self.undo_click_button.configure(state=after_1st_click_state)
        self.reset_clicks_button.configure(state=after_1st_click_state)
        self.zoomin_options_frame.set_frame_state(before_1st_click_state)
        self.brs_options_frame.set_frame_state(before_1st_click_state)

        if self.state["brs_mode"].get() == "NoBRS":
            self.net_clicks_entry.configure(state=tk.DISABLED)
            self.net_clicks_label.configure(state=tk.DISABLED)
            self.lbfgs_iters_entry.configure(state=tk.DISABLED)
            self.lbfgs_iters_label.configure(state=tk.DISABLED)

    def _check_entry(self, widget):
        all_checked = True
        if widget.winfo_children is not None:
            for w in widget.winfo_children():
                all_checked = all_checked and self._check_entry(w)

        if getattr(widget, "_check_bounds", None) is not None:
            all_checked = all_checked and widget._check_bounds(
                widget.get(), "-1"
            )

        return all_checked


COMMAND_DESCRIPTION = _("Interactively annotate a dataset")


def command(subparser):
    if not torch.cuda.is_available():
        logger.warn(_("CUDA support not detected. Stuff will be way slower and sluggish!"))
    subparser.add_argument("input", type=Path)
    subparser.add_argument("output", type=Path)
    subparser.add_argument(
        "-d",
        "--device",
        dest="device",
        type=torch.device,
        default="cuda" if torch.cuda.is_available() else "cpu",
    )
    subparser.add_argument(
        "-c", "--classes", dest="classes", type=str, required=True, nargs="+"
    )
    subparser.add_argument(
        "-w", "--checkpoint", dest="checkpoint", type=Path
    )
    subparser.add_argument(
        "-s",
        "--seed",
        dest="seed",
        type=Path,
        help=_("Folder with pre-defined segmentations to be adjusted"),
        # help="Pasta com segmentações pré definidas para serem ajustadas",
    )

    def handle(args):
        logger.debug(_("classes: ") + ", ".join(args.classes))

        assert args.input.is_dir(), _("Origin should be a folder")
        # assert args.input.is_dir(), _("Origem precisa ser uma pasta")
        assert not args.output.exists() or args.output.is_dir(), _(
            "Destination should be a folder, if it isn't exist it will be created"
        )
        # ), "Destino precisa ser uma pasta, se não existe vai ser criada"

        if args.checkpoint is None:
            logger.info(_("Checkpoint not specified, using default..."))
            args.checkpoint = get_default_weight()

        torch.backends.cudnn.determistic = True

        checkpoint_path = find_checkpoint(
            args.checkpoint.parent, args.checkpoint.name
        )
        model = load_is_model(checkpoint_path, args.device, cpu_dist_maps=True)

        def look_for_tasks():
            for class_name in args.classes:
                for image in args.input.glob("*.jpg"):
                    image_name = image.name
                    output_dir = args.output / image_name
                    output_dir.mkdir(exist_ok=True, parents=True)
                    mask_output = output_dir / f"{class_name}.png"
                    if mask_output.exists():
                        continue
                    yield edict(
                        dict(
                            output_dir=output_dir,
                            image=image,
                            name=image_name,
                            mask_output=mask_output,
                            class_name=class_name,
                            seed_mask=args.seed
                            / image_name
                            / f"{class_name}.png"
                            if args.seed is not None
                            else None,
                        )
                    )

        tasks = look_for_tasks()

        root = tk.Tk()
        root.minsize(960, 480)

        app_args = edict(dict(limit_longest_size=400, device=args.device))

        app = InteractiveDemoApp(root, app_args, model, tasks)
        root.deiconify()
        app.mainloop()

    return handle



import math
import random
from functools import lru_cache

import cv2
import numpy as np

from .sample import DSample


class BasePointSampler:
    def __init__(self):
        self._selected_mask = None
        self._selected_masks = None

    def sample_object(self, sample: DSample):
        raise NotImplementedError

    def sample_points(self):
        raise NotImplementedError

    @property
    def selected_mask(self):
        assert self._selected_mask is not None
        return self._selected_mask

    @selected_mask.setter
    def selected_mask(self, mask):
        self._selected_mask = mask[np.newaxis, :].astype(np.float32)


class MultiPointSampler(BasePointSampler):
    def __init__(
        self,
        max_num_points,
        prob_gamma=0.7,
        expand_ratio=0.1,
        positive_erode_prob=0.9,
        positive_erode_iters=3,
        negative_bg_prob=0.1,
        negative_other_prob=0.4,
        negative_border_prob=0.5,
        merge_objects_prob=0.0,
        max_num_merged_objects=2,
        use_hierarchy=False,
        soft_targets=False,
        first_click_center=False,
        only_one_first_click=False,
        sfc_inner_k=1.7,
        sfc_full_inner_prob=0.0,
    ):
        super().__init__()
        self.max_num_points = max_num_points
        self.expand_ratio = expand_ratio
        self.positive_erode_prob = positive_erode_prob
        self.positive_erode_iters = positive_erode_iters
        self.merge_objects_prob = merge_objects_prob
        self.use_hierarchy = use_hierarchy
        self.soft_targets = soft_targets
        self.first_click_center = first_click_center
        self.only_one_first_click = only_one_first_click
        self.sfc_inner_k = sfc_inner_k
        self.sfc_full_inner_prob = sfc_full_inner_prob

        if max_num_merged_objects == -1:
            max_num_merged_objects = max_num_points
        self.max_num_merged_objects = max_num_merged_objects

        self.neg_strategies = ["bg", "other", "border"]
        self.neg_strategies_prob = [
            negative_bg_prob,
            negative_other_prob,
            negative_border_prob,
        ]
        assert math.isclose(sum(self.neg_strategies_prob), 1.0)

        self._pos_probs = generate_probs(max_num_points, gamma=prob_gamma)
        self._neg_probs = generate_probs(max_num_points + 1, gamma=prob_gamma)
        self._neg_masks = None

    def sample_object(self, sample: DSample):
        if len(sample) == 0:
            bg_mask = sample.get_background_mask()
            self.selected_mask = np.zeros_like(bg_mask, dtype=np.float32)
            self._selected_masks = [[]]
            self._neg_masks = {
                strategy: bg_mask for strategy in self.neg_strategies
            }
            self._neg_masks["required"] = []
            return

        gt_mask, pos_masks, neg_masks = self._sample_mask(sample)
        binary_gt_mask = gt_mask > 0.5 if self.soft_targets else gt_mask > 0

        self.selected_mask = gt_mask
        self._selected_masks = pos_masks

        neg_mask_bg = np.logical_not(binary_gt_mask)
        neg_mask_border = self._get_border_mask(binary_gt_mask)
        if len(sample) <= len(self._selected_masks):
            neg_mask_other = neg_mask_bg
        else:
            neg_mask_other = np.logical_and(
                np.logical_not(sample.get_background_mask()),
                np.logical_not(binary_gt_mask),
            )

        self._neg_masks = {
            "bg": neg_mask_bg,
            "other": neg_mask_other,
            "border": neg_mask_border,
            "required": neg_masks,
        }

    def _sample_mask(self, sample: DSample):
        root_obj_ids = sample.root_objects

        if len(root_obj_ids) > 1 and random.random() < self.merge_objects_prob:
            max_selected_objects = min(
                len(root_obj_ids), self.max_num_merged_objects
            )
            num_selected_objects = np.random.randint(
                2, max_selected_objects + 1
            )
            random_ids = random.sample(root_obj_ids, num_selected_objects)
        else:
            random_ids = [random.choice(root_obj_ids)]

        gt_mask = None
        pos_segments = []
        neg_segments = []
        for obj_id in random_ids:
            (
                obj_gt_mask,
                obj_pos_segments,
                obj_neg_segments,
            ) = self._sample_from_masks_layer(obj_id, sample)
            if gt_mask is None:
                gt_mask = obj_gt_mask
            else:
                gt_mask = np.maximum(gt_mask, obj_gt_mask)

            pos_segments.extend(obj_pos_segments)
            neg_segments.extend(obj_neg_segments)

        pos_masks = [self._positive_erode(x) for x in pos_segments]
        neg_masks = [self._positive_erode(x) for x in neg_segments]

        return gt_mask, pos_masks, neg_masks

    def _sample_from_masks_layer(self, obj_id, sample: DSample):
        objs_tree = sample._objects

        if not self.use_hierarchy:
            node_mask = sample.get_object_mask(obj_id)
            gt_mask = (
                sample.get_soft_object_mask(obj_id)
                if self.soft_targets
                else node_mask
            )
            return gt_mask, [node_mask], []

        def _select_node(node_id):
            node_info = objs_tree[node_id]
            if not node_info["children"] or random.random() < 0.5:
                return node_id
            return _select_node(random.choice(node_info["children"]))

        selected_node = _select_node(obj_id)
        node_info = objs_tree[selected_node]
        node_mask = sample.get_object_mask(selected_node)
        gt_mask = (
            sample.get_soft_object_mask(selected_node)
            if self.soft_targets
            else node_mask
        )
        pos_mask = node_mask.copy()

        negative_segments = []
        if (
            node_info["parent"] is not None
            and node_info["parent"] in objs_tree
        ):
            parent_mask = sample.get_object_mask(node_info["parent"])
            negative_segments.append(
                np.logical_and(parent_mask, np.logical_not(node_mask))
            )

        for child_id in node_info["children"]:
            if objs_tree[child_id]["area"] / node_info["area"] < 0.10:
                child_mask = sample.get_object_mask(child_id)
                pos_mask = np.logical_and(pos_mask, np.logical_not(child_mask))

        if node_info["children"]:
            max_disabled_children = min(len(node_info["children"]), 3)
            num_disabled_children = np.random.randint(
                0, max_disabled_children + 1
            )
            disabled_children = random.sample(
                node_info["children"], num_disabled_children
            )

            for child_id in disabled_children:
                child_mask = sample.get_object_mask(child_id)
                pos_mask = np.logical_and(pos_mask, np.logical_not(child_mask))
                if self.soft_targets:
                    soft_child_mask = sample.get_soft_object_mask(child_id)
                    gt_mask = np.minimum(gt_mask, 1.0 - soft_child_mask)
                else:
                    gt_mask = np.logical_and(
                        gt_mask, np.logical_not(child_mask)
                    )
                negative_segments.append(child_mask)

        return gt_mask, [pos_mask], negative_segments

    def sample_points(self):
        assert self._selected_mask is not None
        pos_points = self._multi_mask_sample_points(
            self._selected_masks,
            is_negative=[False] * len(self._selected_masks),
            with_first_click=self.first_click_center,
        )

        neg_strategy = [
            (self._neg_masks[k], prob)
            for k, prob in zip(self.neg_strategies, self.neg_strategies_prob)
        ]
        neg_masks = self._neg_masks["required"] + [neg_strategy]
        neg_points = self._multi_mask_sample_points(
            neg_masks,
            is_negative=[False] * len(self._neg_masks["required"]) + [True],
        )

        return pos_points + neg_points

    def _multi_mask_sample_points(
        self, selected_masks, is_negative, with_first_click=False
    ):
        selected_masks = selected_masks[: self.max_num_points]

        each_obj_points = [
            self._sample_points(
                mask,
                is_negative=is_negative[i],
                with_first_click=with_first_click,
            )
            for i, mask in enumerate(selected_masks)
        ]
        each_obj_points = [x for x in each_obj_points if len(x) > 0]

        points = []
        if len(each_obj_points) == 1:
            points = each_obj_points[0]
        elif len(each_obj_points) > 1:
            if self.only_one_first_click:
                each_obj_points = each_obj_points[:1]

            points = [obj_points[0] for obj_points in each_obj_points]

            aggregated_masks_with_prob = []
            for indx, x in enumerate(selected_masks):
                if (
                    isinstance(x, (list, tuple))
                    and x
                    and isinstance(x[0], (list, tuple))
                ):
                    for t, prob in x:
                        aggregated_masks_with_prob.append(
                            (t, prob / len(selected_masks))
                        )
                else:
                    aggregated_masks_with_prob.append(
                        (x, 1.0 / len(selected_masks))
                    )

            other_points_union = self._sample_points(
                aggregated_masks_with_prob, is_negative=True
            )
            if len(other_points_union) + len(points) <= self.max_num_points:
                points.extend(other_points_union)
            else:
                points.extend(
                    random.sample(
                        other_points_union, self.max_num_points - len(points)
                    )
                )

        if len(points) < self.max_num_points:
            points.extend([(-1, -1, -1)] * (self.max_num_points - len(points)))

        return points

    def _sample_points(self, mask, is_negative=False, with_first_click=False):
        if is_negative:
            num_points = np.random.choice(
                np.arange(self.max_num_points + 1), p=self._neg_probs
            )
        else:
            num_points = 1 + np.random.choice(
                np.arange(self.max_num_points), p=self._pos_probs
            )

        indices_probs = None
        if isinstance(mask, (list, tuple)):
            indices_probs = [x[1] for x in mask]
            indices = [(np.argwhere(x), prob) for x, prob in mask]
            if indices_probs:
                assert math.isclose(sum(indices_probs), 1.0)
        else:
            indices = np.argwhere(mask)

        points = []
        for j in range(num_points):
            first_click = with_first_click and j == 0 and indices_probs is None

            if first_click:
                point_indices = get_point_candidates(
                    mask,
                    k=self.sfc_inner_k,
                    full_prob=self.sfc_full_inner_prob,
                )
            elif indices_probs:
                point_indices_indx = np.random.choice(
                    np.arange(len(indices)), p=indices_probs
                )
                point_indices = indices[point_indices_indx][0]
            else:
                point_indices = indices

            num_indices = len(point_indices)
            if num_indices > 0:
                point_indx = 0 if first_click else 100
                click = point_indices[
                    np.random.randint(0, num_indices)
                ].tolist() + [point_indx]
                points.append(click)

        return points

    def _positive_erode(self, mask):
        if random.random() > self.positive_erode_prob:
            return mask

        kernel = np.ones((3, 3), np.uint8)
        eroded_mask = cv2.erode(
            mask.astype(np.uint8), kernel, iterations=self.positive_erode_iters
        ).astype(bool)

        if eroded_mask.sum() > 10:
            return eroded_mask
        else:
            return mask

    def _get_border_mask(self, mask):
        expand_r = int(np.ceil(self.expand_ratio * np.sqrt(mask.sum())))
        kernel = np.ones((3, 3), np.uint8)
        expanded_mask = cv2.dilate(
            mask.astype(np.uint8), kernel, iterations=expand_r
        )
        expanded_mask[mask.astype(bool)] = 0
        return expanded_mask


@lru_cache(maxsize=None)
def generate_probs(max_num_points, gamma):
    probs = []
    last_value = 1
    for i in range(max_num_points):
        probs.append(last_value)
        last_value *= gamma

    probs = np.array(probs)
    probs /= probs.sum()

    return probs


def get_point_candidates(obj_mask, k=1.7, full_prob=0.0):
    if full_prob > 0 and random.random() < full_prob:
        return obj_mask

    padded_mask = np.pad(obj_mask, ((1, 1), (1, 1)), "constant")

    dt = cv2.distanceTransform(padded_mask.astype(np.uint8), cv2.DIST_L2, 0)[
        1:-1, 1:-1
    ]
    if k > 0:
        inner_mask = dt > dt.max() / k
        return np.argwhere(inner_mask)
    else:
        prob_map = dt.flatten()
        prob_map /= max(prob_map.sum(), 1e-6)
        click_indx = np.random.choice(len(prob_map), p=prob_map)
        click_coords = np.unravel_index(click_indx, dt.shape)
        return np.array([click_coords])



import torch

from .resnetv1b import resnet34_v1b, resnet50_v1s, resnet101_v1s, resnet152_v1s


class ResNetBackbone(torch.nn.Module):
    def __init__(
        self, backbone="resnet50", pretrained_base=True, dilated=True, **kwargs
    ):
        super(ResNetBackbone, self).__init__()

        if backbone == "resnet34":
            pretrained = resnet34_v1b(
                pretrained=pretrained_base, dilated=dilated, **kwargs
            )
        elif backbone == "resnet50":
            pretrained = resnet50_v1s(
                pretrained=pretrained_base, dilated=dilated, **kwargs
            )
        elif backbone == "resnet101":
            pretrained = resnet101_v1s(
                pretrained=pretrained_base, dilated=dilated, **kwargs
            )
        elif backbone == "resnet152":
            pretrained = resnet152_v1s(
                pretrained=pretrained_base, dilated=dilated, **kwargs
            )
        else:
            raise RuntimeError(
                _("unknown backbone: {backbone}").format(backbone=backbone)
            )

        self.conv1 = pretrained.conv1
        self.bn1 = pretrained.bn1
        self.relu = pretrained.relu
        self.maxpool = pretrained.maxpool
        self.layer1 = pretrained.layer1
        self.layer2 = pretrained.layer2
        self.layer3 = pretrained.layer3
        self.layer4 = pretrained.layer4

    def forward(self, x, additional_features=None):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        if additional_features is not None:
            x = x + torch.nn.functional.pad(
                additional_features,
                [0, 0, 0, 0, 0, x.size(1) - additional_features.size(1)],
                mode="constant",
                value=0,
            )
        x = self.maxpool(x)
        c1 = self.layer1(x)
        c2 = self.layer2(c1)
        c3 = self.layer3(c2)
        c4 = self.layer4(c3)

        return c1, c2, c3, c4



import logging
import os
from pathlib import Path

import torch

from ritm_annotation.utils.env import load_cfg_from_env
from ritm_annotation.utils.exp import init_experiment
from ritm_annotation.utils.misc import load_module

from .dataset import (
    AnnotationDataset,
    get_points_sampler,
    get_train_augmentator,
    get_val_augmentator,
)

logger = logging.getLogger(__name__)


COMMAND_DESCRIPTION = _(
    "Run finetune trains using a dataset in the format the annotator generates"
)


def command(parser):
    parser.add_argument(
        "model_path", type=Path, help=_("Path to the model script.")
    )
    parser.add_argument(
        "images_path", type=Path, help=_("Path to the dataset images.")
    )
    parser.add_argument(
        "masks_path", type=Path, help=_("Path to the dataset masks.")
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="experiment_path",
        type=Path,
        default=Path("."),
        help=_("Where to store experiment data"),
    )

    parser.add_argument(
        "-n",
        "--num-epochs",
        dest="num_epochs",
        type=int,
        help=_("Amount of epochs"),
    )

    parser.add_argument(
        "--exp-name",
        type=str,
        default="",
        help=_(
            "Here you can specify the name of the experiment. It will be added as a suffix to the experiment folder."
        ),
    )  # noqa:E501

    parser.add_argument(
        "--workers",
        type=int,
        default=4,
        metavar="N",
        help=_("Dataloader threads."),
    )

    parser.add_argument(
        "--batch-size",
        "-b",
        dest="batch_size",
        type=int,
        default=-1,
        help=_(
            "You can override model batch size by specify positive number."
        ),
    )  # noqa:E501

    parser.add_argument(
        "--ngpus",
        type=int,
        default=1,
        help=_(
            'Number of GPUs. If you only specify "--gpus" argument, the ngpus value will be calculated automatically. You should use either this argument or "--gpus".'
        ),  # noqa:E501
    )  # noqa:E501

    parser.add_argument(
        "--gpus",
        type=str,
        default="",
        required=False,
        help=_(
            'Ids of used GPUs. You should use either this argument or "--ngpus".'
        ),  # noqa: E501
    )  # noqa:E501

    parser.add_argument(
        "--resume-exp",
        type=str,
        default=None,
        help=_(
            'The prefix of the name of the experiment to be continued. If you use this field, you must specify the "--resume-prefix" argument.'
        ),  # noqa:E501
    )  # noqa:E501

    parser.add_argument(
        "--resume-prefix",
        type=str,
        default="latest",
        help=_("The prefix of the name of the checkpoint to be loaded."),
    )  # noqa:E501

    parser.add_argument(
        "--start-epoch",
        type=int,
        default=0,
        help=_(
            "The number of the starting epoch from which training will continue. (it is important for correct logging and learning rate)"
        ),
    )  # noqa:E501

    parser.add_argument(
        "--weights",
        type=str,
        default=None,
        help=_(
            "Model weights will be loaded from the specified path if you use this argument."
        ),  # noqa: E501
    )  # noqa:E501

    parser.add_argument(
        "--temp-model-path",
        type=str,
        default="",
        help=_("Do not use this argument (for internal purposes)."),
    )  # noqa:E501

    parser.add_argument(
        "--max-bigger-dimension",
        dest="max_bigger_dimension",
        type=int,
        help=_(
            "Resize the input dataset so the bigger dimension is this value"
        ),
    )

    parser.add_argument("--local_rank", type=int, default=0)

    def handle(args):
        model_path = Path(args.model_path)
        if args.temp_model_path != "":
            logger.debug(
                _(
                    "Falling back to temp_model_path because model_path wasn't specified"
                )  # noqa:E501
            )
            model_path = Path(args.temp_model_path)
        if not model_path.is_absolute():
            model_path = (
                Path(__file__).parent.parent.parent / "models" / model_path
            )
        model_path = model_path.resolve()
        args.model_path = model_path
        logger.debug(
            _("Final model path: '{model_path}'").format(model_path=model_path)
        )

        model_script = load_module(model_path)

        model_base_name = model_script.__dict__.get("MODEL_NAME")

        args.distributed = "WORLD_SIZE" in os.environ
        cfg = init_experiment(args, model_base_name)
        cfg.weights = args.weights
        cfg = load_cfg_from_env(cfg, os.environ)

        torch.backends.cudnn.benchmark = True
        torch.multiprocessing.set_sharing_strategy("file_system")
        logger.debug(_("Basic validations passed"))
        model, model_cfg = model_script.init_model(cfg)

        train_augmentator = get_train_augmentator(model_cfg)
        val_augmentator = get_val_augmentator(model_cfg)

        trainer = model_script.get_trainer(
            model, cfg, model_cfg, no_dataset=True
        )
        points_sampler = get_points_sampler(model_cfg)
        dataset_common_args = dict(
            images_path=args.images_path,
            masks_path=args.masks_path,
            keep_background_prob=0.05,
            min_object_area=1000,
            points_sampler=points_sampler,
        )
        if args.max_bigger_dimension is not None:
            dataset_common_args[
                "max_bigger_dimension"
            ] = args.max_bigger_dimension
        trainer.trainset = AnnotationDataset(
            split="train",
            augmentator=train_augmentator,
            **dataset_common_args,
        )
        trainer.valset = AnnotationDataset(
            split="val",
            augmentator=val_augmentator,
            **dataset_common_args,
        )
        if args.num_epochs is None:
            args.num_epochs = model_cfg.default_num_epochs
        trainer._before_needed_hook()
        assert trainer.lr_scheduler is not None
        trainer.run(num_epochs=args.num_epochs)

    return handle



import pickle
import random

import numpy as np
import torch
from torchvision import transforms

from .points_sampler import MultiPointSampler
from .sample import DSample


class ISDataset(torch.utils.data.dataset.Dataset):
    def __init__(
        self,
        augmentator=None,
        points_sampler=MultiPointSampler(max_num_points=12),
        min_object_area=0,
        keep_background_prob=0.0,
        with_image_info=False,
        samples_scores_path=None,
        samples_scores_gamma=1.0,
        epoch_len=-1,
        dry_run=False,
    ):
        super(ISDataset, self).__init__()
        self.epoch_len = epoch_len
        self.augmentator = augmentator
        self.min_object_area = min_object_area
        self.keep_background_prob = keep_background_prob
        self.points_sampler = points_sampler
        self.with_image_info = with_image_info
        self.samples_precomputed_scores = self._load_samples_scores(
            samples_scores_path, samples_scores_gamma, dry_run=dry_run
        )
        self.to_tensor = transforms.ToTensor()

        self.dataset_samples = None
        self.dry_run = dry_run

    def __getitem__(self, index):
        if self.samples_precomputed_scores is not None:
            index = np.random.choice(
                self.samples_precomputed_scores["indices"],
                p=self.samples_precomputed_scores["probs"],
            )
        else:
            if self.epoch_len > 0:
                index = random.randrange(0, len(self.dataset_samples))

        sample = self.get_sample(index)
        sample = self.augment_sample(sample)
        sample.remove_small_objects(self.min_object_area)

        self.points_sampler.sample_object(sample)
        points = np.array(self.points_sampler.sample_points())
        mask = self.points_sampler.selected_mask

        output = {
            "images": self.to_tensor(sample.image),
            "points": points.astype(np.float32),
            "instances": mask,
        }

        if self.with_image_info:
            output["image_info"] = sample.sample_id

        return output

    def augment_sample(self, sample) -> DSample:
        if self.augmentator is None:
            return sample

        valid_augmentation = False
        while not valid_augmentation:
            sample.augment(self.augmentator)
            keep_sample = (
                self.keep_background_prob < 0.0
                or random.random() < self.keep_background_prob
            )
            valid_augmentation = len(sample) > 0 or keep_sample

        return sample

    def get_sample(self, index) -> DSample:
        raise NotImplementedError

    def __len__(self):
        if self.epoch_len > 0:
            return self.epoch_len
        else:
            return self.get_samples_number()

    def get_samples_number(self):
        return len(self.dataset_samples)

    @staticmethod
    def _load_samples_scores(
        samples_scores_path, samples_scores_gamma, dry_run=False
    ):
        if samples_scores_path is None or (not dry_run):
            return None
        with open(samples_scores_path, "rb") as f:
            images_scores = pickle.load(f)

        probs = np.array(
            [(1.0 - x[2]) ** samples_scores_gamma for x in images_scores]
        )
        probs /= probs.sum()
        samples_scores = {
            "indices": [x[0] for x in images_scores],
            "probs": probs,
        }
        print(
            _("Loaded {num_weights} weights with gamma={gamma}").format(
                num_weights=len(probs), gamma=samples_scores_gamma
            )
        )
        return samples_scores



import logging
import os
import pprint
import random
from collections import defaultdict
from copy import deepcopy

import cv2
import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from ritm_annotation.utils.distributed import (
    get_dp_wrapper,
    get_sampler,
    reduce_loss_dict,
)
from ritm_annotation.utils.log import SummaryWriterAvg, TqdmToLogger
from ritm_annotation.utils.misc import save_checkpoint
from ritm_annotation.utils.serialization import get_config_repr
from ritm_annotation.utils.vis import draw_points, draw_probmap

from .optimizer import get_optimizer

logger = logging.getLogger(__name__)


class ISTrainer(object):
    def __init__(
        self,
        model,
        cfg,
        model_cfg,
        loss_cfg,
        trainset=None,
        valset=None,
        optimizer="adam",
        optimizer_params=None,
        image_dump_interval=200,
        checkpoint_interval=10,
        tb_dump_period=25,
        max_interactive_points=0,
        lr_scheduler=None,
        metrics=None,
        additional_val_metrics=None,
        net_inputs=("images", "points"),
        max_num_next_clicks=0,
        click_models=None,
        prev_mask_drop_prob=0.0,
        dry_run=False,
    ):
        self._ran_before_needed = False
        self.cfg = cfg
        self.model_cfg = model_cfg
        self.max_interactive_points = max_interactive_points
        self.loss_cfg = loss_cfg
        self.val_loss_cfg = deepcopy(loss_cfg)
        self.tb_dump_period = tb_dump_period
        self.net_inputs = net_inputs
        self.max_num_next_clicks = max_num_next_clicks

        self.click_models = click_models
        self.prev_mask_drop_prob = prev_mask_drop_prob

        if cfg.distributed:
            cfg.batch_size //= cfg.ngpus
            cfg.val_batch_size //= cfg.ngpus

        if metrics is None:
            metrics = []
        self.train_metrics = metrics
        self.val_metrics = deepcopy(metrics)
        if additional_val_metrics is not None:
            self.val_metrics.extend(additional_val_metrics)

        self.checkpoint_interval = checkpoint_interval
        self.image_dump_interval = image_dump_interval
        self.task_prefix = ""
        self.sw = None

        self.dry_run = dry_run
        self.trainset = trainset
        self.valset = valset
        self.device = self.cfg.device
        self.lr_scheduler_fn = lr_scheduler
        self.model = model
        self.optimizer = optimizer
        self.optimizer_params = optimizer_params

        self.tqdm_out = TqdmToLogger(logger, level=logging.INFO)

    def _before_needed_hook(self):
        if self._ran_before_needed:
            return
        self._ran_before_needed = True

        logger.debug(_("Looking for basic assumptions"))
        assert self.trainset is not None, _("Missing train dataset")
        assert self.valset is not None, _("Missing validation dataset")

        logger.info(
            _(
                "Dataset of {num_samples} samples was loaded for training."
            ).format(
                num_samples=self.trainset.get_samples_number()
            )  # noqa: E501
        )
        logger.info(
            _(
                "Dataset of {num_samples} samples was loaded for validation."
            ).format(
                num_samples=self.valset.get_samples_number()
            )  # noqa: E501
        )

        logger.debug(_("Setting up dataloder"))
        self.train_data = DataLoader(
            self.trainset,
            self.cfg.batch_size,
            sampler=get_sampler(
                self.trainset, shuffle=True, distributed=self.cfg.distributed
            )
            if not self.dry_run
            else None,
            drop_last=True,
            pin_memory=True,
            num_workers=self.cfg.workers,
            persistent_workers=self.cfg.workers > 0,
        )

        self.val_data = DataLoader(
            self.valset,
            self.cfg.val_batch_size,
            sampler=get_sampler(
                self.valset, shuffle=False, distributed=self.cfg.distributed
            )
            if not self.dry_run
            else None,
            drop_last=True,
            pin_memory=True,
            num_workers=self.cfg.workers,
            persistent_workers=self.cfg.workers > 0,
        )
        logger.debug(_("Initializing model"))
        self.model = self._load_weights(self.model)
        self.optim = get_optimizer(
            self.model, self.optimizer, self.optimizer_params
        )

        if self.cfg.multi_gpu:
            self.model = get_dp_wrapper(self.cfg.distributed)(
                self.model,
                device_ids=self.cfg.gpu_ids,
                output_device=self.cfg.gpu_ids[0],
            )

        if self.is_master:
            logger.info(self.model)
            logger.info(get_config_repr(self.model._config))

        self.net = self.model.to(self.device)
        self.lr = self.optimizer_params["lr"]

        if self.lr_scheduler_fn is not None:
            self.lr_scheduler = self.lr_scheduler_fn(optimizer=self.optim)
            if self.cfg.start_epoch > 0:
                for __ in range(self.cfg.start_epoch):
                    self.lr_scheduler.step()

        logger.debug(_("Initializing click models"))

        if self.click_models is not None:
            for click_model in self.click_models:
                for param in click_model.parameters():
                    param.requires_grad = False
                click_model.to(self.device)
                click_model.eval()

        logger.debug(
            _("First train batch: {batch}").format(
                batch=repr(next(iter(self.train_data)))
            )
        )
        logger.debug(
            _("First evaluation batch: {batch}").format(
                batch=repr(next(iter(self.val_data)))
            )
        )

        logger.info(_("Run experiment with config:"))
        logger.info(pprint.pformat(self.cfg, indent=4))

    def run(self, num_epochs, start_epoch=None, validation=True):
        self._before_needed_hook()
        if start_epoch is None:
            start_epoch = self.cfg.start_epoch

        logger.info(
            _("Starting Epoch: {start_epoch}").format(start_epoch=start_epoch)
        )
        logger.info(
            _("Total Epochs: {num_epochs}").format(num_epochs=num_epochs)
        )
        for epoch in range(start_epoch, num_epochs):
            self.training(epoch)
            if validation:
                self.validation(epoch)

    def training(self, epoch):
        self._before_needed_hook()
        if self.sw is None and self.is_master:
            self.sw = SummaryWriterAvg(
                log_dir=str(self.cfg.LOGS_PATH),
                flush_secs=10,
                dump_period=self.tb_dump_period,
            )

        if self.cfg.distributed:
            self.train_data.sampler.set_epoch(epoch)

        log_prefix = "Train" + self.task_prefix.capitalize()
        tbar = (
            tqdm(self.train_data, file=self.tqdm_out, ncols=100)
            if self.is_master
            else self.train_data
        )

        for metric in self.train_metrics:
            metric.reset_epoch_stats()

        self.net.train()
        train_loss = 0.0
        for i, batch_data in enumerate(tbar):
            global_step = epoch * len(self.train_data) + i

            (
                loss,
                losses_logging,
                splitted_batch_data,
                outputs,
            ) = self.batch_forward(batch_data)

            self.optim.zero_grad()
            loss.backward()
            self.optim.step()

            losses_logging["overall"] = loss
            reduce_loss_dict(losses_logging)

            train_loss += losses_logging["overall"].item()

            if self.is_master:
                for loss_name, loss_value in losses_logging.items():
                    self.sw.add_scalar(
                        tag=f"{log_prefix}Losses/{loss_name}",
                        value=loss_value.item(),
                        global_step=global_step,
                    )

                for k, v in self.loss_cfg.items():
                    if (
                        "_loss" in k
                        and hasattr(v, "log_states")
                        and self.loss_cfg.get(k + "_weight", 0.0) > 0
                    ):
                        v.log_states(
                            self.sw, f"{log_prefix}Losses/{k}", global_step
                        )

                if (
                    self.image_dump_interval > 0
                    and global_step % self.image_dump_interval == 0
                ):
                    self.save_visualization(
                        splitted_batch_data,
                        outputs,
                        global_step,
                        prefix="train",
                    )

                self.sw.add_scalar(
                    tag=f"{log_prefix}States/learning_rate",
                    value=self.lr
                    if not hasattr(self, "lr_scheduler")
                    else self.lr_scheduler.get_lr()[-1],
                    global_step=global_step,
                )

                tbar.set_description(
                    _("Epoch {epoch}, training loss {train_loss:.8f}").format(
                        epoch=epoch, train_loss=train_loss / (i + 1)
                    )
                )
                for metric in self.train_metrics:
                    metric.log_states(
                        self.sw,
                        f"{log_prefix}Metrics/{metric.name}",
                        global_step,
                    )

        if self.is_master:
            for metric in self.train_metrics:
                self.sw.add_scalar(
                    tag=f"{log_prefix}Metrics/{metric.name}",
                    value=metric.get_epoch_value(),
                    global_step=epoch,
                    disable_avg=True,
                )

            save_checkpoint(
                self.net,
                self.cfg.CHECKPOINTS_PATH,
                prefix=self.task_prefix,
                epoch=None,
                multi_gpu=self.cfg.multi_gpu,
            )

            if isinstance(self.checkpoint_interval, (list, tuple)):
                checkpoint_interval = [
                    x for x in self.checkpoint_interval if x[0] <= epoch
                ][-1][1]
            else:
                checkpoint_interval = self.checkpoint_interval

            if epoch % checkpoint_interval == 0:
                save_checkpoint(
                    self.net,
                    self.cfg.CHECKPOINTS_PATH,
                    prefix=self.task_prefix,
                    epoch=epoch,
                    multi_gpu=self.cfg.multi_gpu,
                )

        if hasattr(self, "lr_scheduler"):
            self.lr_scheduler.step()

    def validation(self, epoch):
        self._before_needed_hook()
        if self.sw is None and self.is_master:
            self.sw = SummaryWriterAvg(
                log_dir=str(self.cfg.LOGS_PATH),
                flush_secs=10,
                dump_period=self.tb_dump_period,
            )

        log_prefix = "Val" + self.task_prefix.capitalize()
        tbar = (
            tqdm(self.val_data, file=self.tqdm_out, ncols=100)
            if self.is_master
            else self.val_data
        )

        for metric in self.val_metrics:
            metric.reset_epoch_stats()

        val_loss = 0
        losses_logging = defaultdict(list)

        self.net.eval()
        for i, batch_data in enumerate(tbar):
            global_step = epoch * len(self.val_data) + i
            (
                loss,
                batch_losses_logging,
                splitted_batch_data,
                outputs,
            ) = self.batch_forward(batch_data, validation=True)

            batch_losses_logging["overall"] = loss
            reduce_loss_dict(batch_losses_logging)
            for loss_name, loss_value in batch_losses_logging.items():
                losses_logging[loss_name].append(loss_value.item())

            val_loss += batch_losses_logging["overall"].item()

            if self.is_master:
                tbar.set_description(
                    _("Epoch {epoch}, validation loss: {val_loss:.4f}").format(
                        epoch=epoch, val_loss=val_loss / (i + 1)
                    )
                )
                for metric in self.val_metrics:
                    metric.log_states(
                        self.sw,
                        f"{log_prefix}Metrics/{metric.name}",
                        global_step,
                    )

        if self.is_master:
            logging.debug(_("Saving epoch level losses"))
            for loss_name, loss_values in losses_logging.items():
                self.sw.add_scalar(
                    tag=f"{log_prefix}Losses/{loss_name}",
                    value=np.array(loss_values).mean(),
                    global_step=epoch,
                    disable_avg=True,
                )

            logging.debug(_("Saving epoch level metrics"))
            for metric in self.val_metrics:
                self.sw.add_scalar(
                    tag=f"{log_prefix}Metrics/{metric.name}",
                    value=metric.get_epoch_value(),
                    global_step=epoch,
                    disable_avg=True,
                )

    def batch_forward(self, batch_data, validation=False):
        self._before_needed_hook()
        metrics = self.val_metrics if validation else self.train_metrics
        losses_logging = dict()

        with torch.set_grad_enabled(not validation):
            batch_data = {k: v.to(self.device) for k, v in batch_data.items()}
            image, gt_mask, points = (
                batch_data["images"],
                batch_data["instances"],
                batch_data["points"],
            )
            # orig_image, orig_gt_mask, orig_points = (
            _, orig_gt_mask, _ = (
                image.clone(),
                gt_mask.clone(),
                points.clone(),
            )

            prev_output = torch.zeros_like(image, dtype=torch.float32)[
                :, :1, :, :
            ]

            last_click_indx = None

            with torch.no_grad():
                num_iters = random.randint(0, self.max_num_next_clicks)

                for click_indx in range(num_iters):
                    last_click_indx = click_indx

                    if not validation:
                        self.net.eval()

                    if self.click_models is None or click_indx >= len(
                        self.click_models
                    ):
                        eval_model = self.net
                    else:
                        eval_model = self.click_models[click_indx]

                    net_input = (
                        torch.cat((image, prev_output), dim=1)
                        if self.net.with_prev_mask
                        else image
                    )
                    prev_output = torch.sigmoid(
                        eval_model(net_input, points)["instances"]
                    )

                    points = get_next_points(
                        prev_output, orig_gt_mask, points, click_indx + 1
                    )

                    if not validation:
                        self.net.train()

                if (
                    self.net.with_prev_mask
                    and self.prev_mask_drop_prob > 0
                    and last_click_indx is not None
                ):
                    zero_mask = (
                        np.random.random(size=prev_output.size(0))
                        < self.prev_mask_drop_prob
                    )
                    prev_output[zero_mask] = torch.zeros_like(
                        prev_output[zero_mask]
                    )

            batch_data["points"] = points

            net_input = (
                torch.cat((image, prev_output), dim=1)
                if self.net.with_prev_mask
                else image
            )
            output = self.net(net_input, points)

            loss = 0.0
            loss = self.add_loss(
                "instance_loss",
                loss,
                losses_logging,
                validation,
                lambda: (output["instances"], batch_data["instances"]),
            )
            loss = self.add_loss(
                "instance_aux_loss",
                loss,
                losses_logging,
                validation,
                lambda: (output["instances_aux"], batch_data["instances"]),
            )

            if self.is_master:
                with torch.no_grad():
                    for m in metrics:
                        m.update(
                            *(output.get(x) for x in m.pred_outputs),
                            *(batch_data[x] for x in m.gt_outputs),
                        )
        return loss, losses_logging, batch_data, output

    def add_loss(
        self,
        loss_name,
        total_loss,
        losses_logging,
        validation,
        lambda_loss_inputs,
    ):
        loss_cfg = self.loss_cfg if not validation else self.val_loss_cfg
        loss_weight = loss_cfg.get(loss_name + "_weight", 0.0)
        if loss_weight > 0.0:
            loss_criterion = loss_cfg.get(loss_name)
            loss = loss_criterion(*lambda_loss_inputs())
            loss = torch.mean(loss)
            losses_logging[loss_name] = loss
            loss = loss_weight * loss
            total_loss = total_loss + loss

        return total_loss

    def save_visualization(
        self, splitted_batch_data, outputs, global_step, prefix
    ):
        output_images_path = self.cfg.VIS_PATH / prefix
        if self.task_prefix:
            output_images_path /= self.task_prefix

        if not output_images_path.exists():
            output_images_path.mkdir(parents=True)
        image_name_prefix = f"{global_step:06d}"

        def _save_image(suffix, image):
            cv2.imwrite(
                str(output_images_path / f"{image_name_prefix}_{suffix}.jpg"),
                image,
                [cv2.IMWRITE_JPEG_QUALITY, 85],
            )

        images = splitted_batch_data["images"]
        points = splitted_batch_data["points"]
        instance_masks = splitted_batch_data["instances"]

        gt_instance_masks = instance_masks.cpu().numpy()
        predicted_instance_masks = (
            torch.sigmoid(outputs["instances"]).detach().cpu().numpy()
        )
        points = points.detach().cpu().numpy()

        image_blob, points = images[0], points[0]
        gt_mask = np.squeeze(gt_instance_masks[0], axis=0)
        predicted_mask = np.squeeze(predicted_instance_masks[0], axis=0)

        image = image_blob.cpu().numpy() * 255
        image = image.transpose((1, 2, 0))

        image_with_points = draw_points(
            image, points[: self.max_interactive_points], (0, 255, 0)
        )
        image_with_points = draw_points(
            image_with_points,
            points[self.max_interactive_points :],
            (0, 0, 255),
        )

        gt_mask[gt_mask < 0] = 0.25
        gt_mask = draw_probmap(gt_mask)
        predicted_mask = draw_probmap(predicted_mask)
        viz_image = np.hstack(
            (image_with_points, gt_mask, predicted_mask)
        ).astype(np.uint8)

        _save_image("instance_segmentation", viz_image[:, :, ::-1])

    def _load_weights(self, net):
        if self.cfg.resume_exp is not None:
            checkpoint_glob = f"{self.cfg.resume_prefix}*.pth"
            checkpoints = list(self.cfg.CHECKPOINTS_PATH.glob(checkpoint_glob))
            assert len(checkpoints) == 1, _(
                "'{checkpoint_glob}' didn't match anything"
            ).format(
                checkpoint_glob=f"{self.cfg.CHECKPOINTS_PATH}/{checkpoint_glob}"
            )

            checkpoint_path = checkpoints[0]
            load_weights(net, str(checkpoint_path))
        elif self.cfg.weights is not None:
            if os.path.isfile(self.cfg.weights):
                load_weights(net, self.cfg.weights)
                self.cfg.weights = None
            else:
                raise RuntimeError(
                    _("=> no checkpoint found at '{weights_dir}'").format(
                        weights_dir=self.cfg.weights
                    )
                )

        return net

    @property
    def is_master(self):
        return self.cfg.local_rank == 0


def get_next_points(pred, gt, points, click_indx, pred_thresh=0.49):
    assert click_indx > 0
    pred = pred.cpu().numpy()[:, 0, :, :]
    gt = gt.cpu().numpy()[:, 0, :, :] > 0.5

    fn_mask = np.logical_and(gt, pred < pred_thresh)
    fp_mask = np.logical_and(np.logical_not(gt), pred > pred_thresh)

    fn_mask = np.pad(fn_mask, ((0, 0), (1, 1), (1, 1)), "constant").astype(
        np.uint8
    )
    fp_mask = np.pad(fp_mask, ((0, 0), (1, 1), (1, 1)), "constant").astype(
        np.uint8
    )
    num_points = points.size(1) // 2
    points = points.clone()

    for bindx in range(fn_mask.shape[0]):
        fn_mask_dt = cv2.distanceTransform(fn_mask[bindx], cv2.DIST_L2, 5)[
            1:-1, 1:-1
        ]
        fp_mask_dt = cv2.distanceTransform(fp_mask[bindx], cv2.DIST_L2, 5)[
            1:-1, 1:-1
        ]

        fn_max_dist = np.max(fn_mask_dt)
        fp_max_dist = np.max(fp_mask_dt)

        is_positive = fn_max_dist > fp_max_dist
        dt = fn_mask_dt if is_positive else fp_mask_dt
        inner_mask = dt > max(fn_max_dist, fp_max_dist) / 2.0
        indices = np.argwhere(inner_mask)
        if len(indices) > 0:
            coords = indices[np.random.randint(0, len(indices))]
            if is_positive:
                points[bindx, num_points - click_indx, 0] = float(coords[0])
                points[bindx, num_points - click_indx, 1] = float(coords[1])
                points[bindx, num_points - click_indx, 2] = float(click_indx)
            else:
                points[bindx, 2 * num_points - click_indx, 0] = float(
                    coords[0]
                )
                points[bindx, 2 * num_points - click_indx, 1] = float(
                    coords[1]
                )
                points[bindx, 2 * num_points - click_indx, 2] = float(
                    click_indx
                )

    return points


def load_weights(model, path_to_weights):
    logger.info(
        _("Loading weights from path: '{path_to_weights}'").format(
            path_to_weights=path_to_weights
        )
    )
    current_state_dict = model.state_dict()
    new_state_dict = torch.load(path_to_weights, map_location="cpu")[
        "state_dict"
    ]
    current_state_dict.update(new_state_dict)
    model.load_state_dict(current_state_dict)



import importlib
import logging

import numpy as np
import torch

from tqdm import tqdm
from pathlib import Path

logger = logging.getLogger(__name__)


def get_dims_with_exclusion(dim, exclude=None):
    dims = list(range(dim))
    if exclude is not None:
        dims.remove(exclude)

    return dims


def save_checkpoint(
    net, checkpoints_path, epoch=None, prefix="", multi_gpu=False
):
    if epoch is None:
        checkpoint_name = "last_checkpoint.pth"
    else:
        checkpoint_name = f"{epoch:03d}.pth"

    if prefix:
        checkpoint_name = f"{prefix}_{checkpoint_name}"

    if not checkpoints_path.exists():
        checkpoints_path.mkdir(parents=True)

    checkpoint_path = checkpoints_path / checkpoint_name
    logger.debug(
        _("Save checkpoint to {checkpoint_path}").format(
            checkpoint_path=str(checkpoint_path)
        )
    )

    net = net.module if multi_gpu else net
    torch.save(
        {"state_dict": net.state_dict(), "config": net._config},
        str(checkpoint_path),
    )


def get_bbox_from_mask(mask):
    rows = np.any(mask, axis=1)
    cols = np.any(mask, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    return rmin, rmax, cmin, cmax


def expand_bbox(bbox, expand_ratio, min_crop_size=None):
    rmin, rmax, cmin, cmax = bbox
    rcenter = 0.5 * (rmin + rmax)
    ccenter = 0.5 * (cmin + cmax)
    height = expand_ratio * (rmax - rmin + 1)
    width = expand_ratio * (cmax - cmin + 1)
    if min_crop_size is not None:
        height = max(height, min_crop_size)
        width = max(width, min_crop_size)

    rmin = int(round(rcenter - 0.5 * height))
    rmax = int(round(rcenter + 0.5 * height))
    cmin = int(round(ccenter - 0.5 * width))
    cmax = int(round(ccenter + 0.5 * width))

    return rmin, rmax, cmin, cmax


def clamp_bbox(bbox, rmin, rmax, cmin, cmax):
    return (
        max(rmin, bbox[0]),
        min(rmax, bbox[1]),
        max(cmin, bbox[2]),
        min(cmax, bbox[3]),
    )


def get_bbox_iou(b1, b2):
    h_iou = get_segments_iou(b1[:2], b2[:2])
    w_iou = get_segments_iou(b1[2:4], b2[2:4])
    return h_iou * w_iou


def get_segments_iou(s1, s2):
    a, b = s1
    c, d = s2
    intersection = max(0, min(b, d) - max(a, c) + 1)
    union = max(1e-6, max(b, d) - min(a, c) + 1)
    return intersection / union


def get_labels_with_sizes(x):
    obj_sizes = np.bincount(x.flatten())
    labels = np.nonzero(obj_sizes)[0].tolist()
    labels = [x for x in labels if x != 0]
    return labels, obj_sizes[labels].tolist()


def ignore_params_then_call(func):
    def ret(*args, **kwargs):
        return func()

    return ret


def load_module(script_path, module_name="module"):
    logger.debug(
        _("Loading module '{module_path}'...").format(module_path=script_path)
    )
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    assert spec is not None, _(
        "Can't import module at '{module_path}'"
    ).format(module_path=script_path)
    model_script = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(model_script)

    return model_script

def get_default_weight():
    from urllib.request import urlopen
    from hashlib import sha256
    OUTPUT_DIR = Path.home() / ".cache" / "ritm_annotation"
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    DEFAULT_MODEL_URL = "https://github.com/SamsungLabs/ritm_interactive_segmentation/releases/download/v1.0/coco_lvis_h18_itermask.pth"
    DEFAULT_MODEL_FILE = OUTPUT_DIR / "coco_lvis_h18_itermask.pth"
    DEFAULT_MODEL_SHA256 = "5f69cfce354d1507e3850bfc39ee7057c8dd27b6a4910d1d2dc724916b9ee32b"
    if DEFAULT_MODEL_FILE.exists():
        return DEFAULT_MODEL_FILE
    try:
        hasher = sha256()
        with urlopen(DEFAULT_MODEL_URL) as req:
            with DEFAULT_MODEL_FILE.open('wb') as f:
                file_size = int(req.headers['Content-Length'])
                print(file_size, type(file_size))
                ops = tqdm(total=file_size, desc=_("Downloading") + f" {DEFAULT_MODEL_URL}")
                while True:
                    buf = req.read(16*1024)
                    if not buf:
                        break
                    hasher.update(buf)
                    f.write(buf)
                    ops.update(len(buf))
        if hasher.hexdigest() != DEFAULT_MODEL_SHA256:
            logger.warning(_("SHA256 of default model is {actual_hash}, expected {expected_hash}").format(
                actual_hash=hasher.hexdigest(),
                expected_hash=DEFAULT_MODEL_SHA256
            ))
        return DEFAULT_MODEL_FILE
    except Exception as e: # se erro deletar o arquivo
        DEFAULT_MODEL_FILE.unlink()
        import traceback
        traceback.print_exc()
        raise e





# flake8: noqa

try:
    from ._get_dist_maps import get_dist_maps
except ImportError:
    import pyximport

    pyximport.install(pyximport=True, language_level=3)
    # noinspection PyUnresolvedReferences
    from ._get_dist_maps import get_dist_maps



# flake8: noqa

from functools import partial

import torch
from albumentations import *
from easydict import EasyDict as edict

from ritm_annotation.data.datasets import *
from ritm_annotation.data.points_sampler import MultiPointSampler
from ritm_annotation.data.transforms import *
from ritm_annotation.engine.trainer import ISTrainer
from ritm_annotation.model import initializer
from ritm_annotation.model.is_deeplab_model import DeeplabModel
from ritm_annotation.model.is_hrnet_model import HRNetModel
from ritm_annotation.model.losses import *
from ritm_annotation.model.metrics import AdaptiveIoU
from ritm_annotation.utils.log import logger



from math import isclose

import numpy as np

from .base import ISDataset


class ComposeDataset(ISDataset):
    def __init__(self, datasets, **kwargs):
        super(ComposeDataset, self).__init__(**kwargs)

        self._datasets = datasets
        self.dataset_samples = []
        for dataset_indx, dataset in enumerate(self._datasets):
            self.dataset_samples.extend(
                [(dataset_indx, i) for i in range(len(dataset))]
            )

    def get_sample(self, index):
        dataset_indx, sample_indx = self.dataset_samples[index]
        return self._datasets[dataset_indx].get_sample(sample_indx)


class ProportionalComposeDataset(ISDataset):
    def __init__(self, datasets, ratios, **kwargs):
        super().__init__(**kwargs)

        assert len(ratios) == len(datasets), _(
            "The number of datasets must match the number of ratios"
        )
        assert isclose(sum(ratios), 1.0), _(
            "The sum of ratios must be equal to 1"
        )

        self._ratios = ratios
        self._datasets = datasets
        self.dataset_samples = []
        for dataset_indx, dataset in enumerate(self._datasets):
            self.dataset_samples.extend(
                [(dataset_indx, i) for i in range(len(dataset))]
            )

    def get_sample(self, index):
        dataset_indx = np.random.choice(len(self._datasets), p=self._ratios)
        sample_indx = np.random.choice(len(self._datasets[dataset_indx]))

        return self._datasets[dataset_indx].get_sample(sample_indx)



import pickle as pkl
from pathlib import Path

import cv2
import numpy as np
from scipy.io import loadmat

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample
from ritm_annotation.utils.misc import (
    get_bbox_from_mask,
    get_labels_with_sizes,
)


class SBDDataset(ISDataset):
    def __init__(
        self,
        dataset_path,
        split="train",
        buggy_mask_thresh=0.08,
        dry_run=False,
        **kwargs,
    ):
        super(SBDDataset, self).__init__(**kwargs)
        assert split in {"train", "val"}

        self.dataset_path = Path(dataset_path)
        self.dataset_split = split
        self._images_path = self.dataset_path / "img"
        self._insts_path = self.dataset_path / "inst"
        self._buggy_objects = dict()
        self._buggy_mask_thresh = buggy_mask_thresh

        if dry_run:
            self.dataset_samples = []
        else:
            with open(self.dataset_path / f"{split}.txt", "r") as f:
                self.dataset_samples = [x.strip() for x in f.readlines()]

    def get_sample(self, index):
        image_name = self.dataset_samples[index]
        image_path = str(self._images_path / f"{image_name}.jpg")
        inst_info_path = str(self._insts_path / f"{image_name}.mat")

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        instances_mask = loadmat(str(inst_info_path))["GTinst"][0][0][
            0
        ].astype(np.int32)
        instances_mask = self.remove_buggy_masks(index, instances_mask)
        instances_ids, _ = get_labels_with_sizes(instances_mask)

        return DSample(
            image, instances_mask, objects_ids=instances_ids, sample_id=index
        )

    def remove_buggy_masks(self, index, instances_mask):
        if self._buggy_mask_thresh > 0.0:
            buggy_image_objects = self._buggy_objects.get(index, None)
            if buggy_image_objects is None:
                buggy_image_objects = []
                instances_ids, _ = get_labels_with_sizes(instances_mask)
                for obj_id in instances_ids:
                    obj_mask = instances_mask == obj_id
                    mask_area = obj_mask.sum()
                    bbox = get_bbox_from_mask(obj_mask)
                    bbox_area = (bbox[1] - bbox[0] + 1) * (
                        bbox[3] - bbox[2] + 1
                    )
                    obj_area_ratio = mask_area / bbox_area
                    if obj_area_ratio < self._buggy_mask_thresh:
                        buggy_image_objects.append(obj_id)

                self._buggy_objects[index] = buggy_image_objects
            for obj_id in buggy_image_objects:
                instances_mask[instances_mask == obj_id] = 0

        return instances_mask


class SBDEvaluationDataset(ISDataset):
    def __init__(self, dataset_path, split="val", dry_run=False, **kwargs):
        super(SBDEvaluationDataset, self).__init__(**kwargs)
        assert split in {"train", "val"}

        self.dataset_path = Path(dataset_path)
        self.dataset_split = split
        self._images_path = self.dataset_path / "img"
        self._insts_path = self.dataset_path / "inst"
        if dry_run:
            self.dataset_samples = []
        else:
            with open(self.dataset_path / f"{split}.txt", "r") as f:
                self.dataset_samples = [x.strip() for x in f.readlines()]
            self.dataset_samples = self.get_sbd_images_and_ids_list()

    def get_sample(self, index) -> DSample:
        image_name, instance_id = self.dataset_samples[index]
        image_path = str(self._images_path / f"{image_name}.jpg")
        inst_info_path = str(self._insts_path / f"{image_name}.mat")

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        instances_mask = loadmat(str(inst_info_path))["GTinst"][0][0][
            0
        ].astype(np.int32)
        instances_mask[instances_mask != instance_id] = 0
        instances_mask[instances_mask > 0] = 1

        return DSample(image, instances_mask, objects_ids=[1], sample_id=index)

    def get_sbd_images_and_ids_list(self):
        pkl_path = (
            self.dataset_path / f"{self.dataset_split}_images_and_ids_list.pkl"
        )

        if pkl_path.exists():
            with open(str(pkl_path), "rb") as fp:
                images_and_ids_list = pkl.load(fp)
        else:
            images_and_ids_list = []

            for sample in self.dataset_samples:
                inst_info_path = str(self._insts_path / f"{sample}.mat")
                instances_mask = loadmat(str(inst_info_path))["GTinst"][0][0][
                    0
                ].astype(np.int32)
                instances_ids, _ = get_labels_with_sizes(instances_mask)

                for instances_id in instances_ids:
                    images_and_ids_list.append((sample, instances_id))

            with open(str(pkl_path), "wb") as fp:
                pkl.dump(images_and_ids_list, fp)

        return images_and_ids_list



import gettext
import logging
from pathlib import Path

logger = logging.getLogger(__name__)

locale_dir = Path(__file__).parent.parent / "i18n"

gettext.install(
    "ritm_annotation",
    localedir=str(locale_dir),
)

gettext.gettext = _

logger.debug(
    _('Loading locale data from "{locale_folder}"').format(
        locale_folder=locale_dir
    )
)



import json
import pickle
import random
from copy import deepcopy
from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample


class CocoLvisDataset(ISDataset):
    def __init__(
        self,
        dataset_path,
        split="train",
        stuff_prob=0.0,
        allow_list_name=None,
        anno_file="hannotation.pickle",
        dry_run=False,
        **kwargs,
    ):
        super(CocoLvisDataset, self).__init__(**kwargs)
        dataset_path = Path(dataset_path)
        self._split_path = dataset_path / split
        self.split = split
        self._images_path = self._split_path / "images"
        self._masks_path = self._split_path / "masks"
        self.stuff_prob = stuff_prob

        if not dry_run:
            with open(self._split_path / anno_file, "rb") as f:
                self.dataset_samples = sorted(pickle.load(f).items())
        else:
            self.dataset_samples = []

        if allow_list_name is not None:
            allow_list_path = self._split_path / allow_list_name
            with open(allow_list_path, "r") as f:
                allow_images_ids = json.load(f)
            allow_images_ids = set(allow_images_ids)

            self.dataset_samples = [
                sample
                for sample in self.dataset_samples
                if sample[0] in allow_images_ids
            ]

    def get_sample(self, index) -> DSample:
        image_id, sample = self.dataset_samples[index]
        image_path = self._images_path / f"{image_id}.jpg"

        image = cv2.imread(str(image_path))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        packed_masks_path = self._masks_path / f"{image_id}.pickle"
        with open(packed_masks_path, "rb") as f:
            encoded_layers, objs_mapping = pickle.load(f)
        layers = [
            cv2.imdecode(x, cv2.IMREAD_UNCHANGED) for x in encoded_layers
        ]
        layers = np.stack(layers, axis=2)

        instances_info = deepcopy(sample["hierarchy"])
        for inst_id, inst_info in list(instances_info.items()):
            if inst_info is None:
                inst_info = {"children": [], "parent": None, "node_level": 0}
                instances_info[inst_id] = inst_info
            inst_info["mapping"] = objs_mapping[inst_id]

        if self.stuff_prob > 0 and random.random() < self.stuff_prob:
            for inst_id in range(
                sample["num_instance_masks"], len(objs_mapping)
            ):
                instances_info[inst_id] = {
                    "mapping": objs_mapping[inst_id],
                    "parent": None,
                    "children": [],
                }
        else:
            for inst_id in range(
                sample["num_instance_masks"], len(objs_mapping)
            ):
                layer_indx, mask_id = objs_mapping[inst_id]
                layers[:, :, layer_indx][  # type: ignore
                    layers[:, :, layer_indx] == mask_id  # type: ignore
                ] = 0

        return DSample(image, layers, objects=instances_info)



from datetime import timedelta
from pathlib import Path

import numpy as np
import torch

from ritm_annotation.data.datasets import (
    BerkeleyDataset,
    DavisDataset,
    GrabCutDataset,
    PascalVocDataset,
    SBDEvaluationDataset,
)
from ritm_annotation.utils.serialization import load_model


def get_time_metrics(all_ious, elapsed_time):
    n_images = len(all_ious)
    n_clicks = sum(map(len, all_ious))

    mean_spc = elapsed_time / n_clicks
    mean_spi = elapsed_time / n_images

    return mean_spc, mean_spi


def load_is_model(checkpoint, device, **kwargs):
    if isinstance(checkpoint, (str, Path)):
        state_dict = torch.load(checkpoint, map_location="cpu")
    else:
        state_dict = checkpoint

    if isinstance(state_dict, list):
        model = load_single_is_model(state_dict[0], device, **kwargs)
        models = [
            load_single_is_model(x, device, **kwargs) for x in state_dict
        ]

        return model, models
    else:
        return load_single_is_model(state_dict, device, **kwargs)


def load_single_is_model(state_dict, device, **kwargs):
    model = load_model(state_dict["config"], **kwargs)
    model.load_state_dict(state_dict["state_dict"], strict=False)

    for param in model.parameters():
        param.requires_grad = False
    model.to(device)
    model.eval()

    return model


def get_dataset(dataset_name, cfg):
    if dataset_name == "GrabCut":
        dataset = GrabCutDataset(cfg.GRABCUT_PATH)
    elif dataset_name == "Berkeley":
        dataset = BerkeleyDataset(cfg.BERKELEY_PATH)
    elif dataset_name == "DAVIS":
        dataset = DavisDataset(cfg.DAVIS_PATH)
    elif dataset_name == "SBD":
        dataset = SBDEvaluationDataset(cfg.SBD_PATH)
    elif dataset_name == "SBD_Train":
        dataset = SBDEvaluationDataset(cfg.SBD_PATH, split="train")
    elif dataset_name == "PascalVOC":
        dataset = PascalVocDataset(cfg.PASCALVOC_PATH, split="test")
    elif dataset_name == "COCO_MVal":
        dataset = DavisDataset(cfg.COCO_MVAL_PATH)
    else:
        dataset = None

    return dataset


def get_iou(gt_mask, pred_mask, ignore_label=-1):
    ignore_gt_mask_inv = gt_mask != ignore_label
    obj_gt_mask = gt_mask == 1

    intersection = np.logical_and(
        np.logical_and(pred_mask, obj_gt_mask), ignore_gt_mask_inv
    ).sum()
    union = np.logical_and(
        np.logical_or(pred_mask, obj_gt_mask), ignore_gt_mask_inv
    ).sum()

    return intersection / union


def compute_noc_metric(all_ious, iou_thrs, max_clicks=20):
    def _get_noc(iou_arr, iou_thr):
        vals = iou_arr >= iou_thr
        return np.argmax(vals) + 1 if np.any(vals) else max_clicks

    noc_list = []
    over_max_list = []
    for iou_thr in iou_thrs:
        scores_arr = np.array(
            [_get_noc(iou_arr, iou_thr) for iou_arr in all_ious], dtype=np.int
        )

        score = scores_arr.mean()
        over_max = (scores_arr == max_clicks).sum()

        noc_list.append(score)
        over_max_list.append(over_max)

    return noc_list, over_max_list


def find_checkpoint(weights_folder, checkpoint_name):
    weights_folder = Path(weights_folder)
    if ":" in checkpoint_name:
        model_name, checkpoint_name = checkpoint_name.split(":")
        models_candidates = [
            x for x in weights_folder.glob(f"{model_name}*") if x.is_dir()
        ]
        assert len(models_candidates) == 1
        model_folder = models_candidates[0]
    else:
        model_folder = weights_folder

    if checkpoint_name.endswith(".pth"):
        if Path(checkpoint_name).exists():
            checkpoint_path = checkpoint_name
        else:
            checkpoint_path = weights_folder / checkpoint_name
    else:
        model_checkpoints = list(model_folder.rglob(f"{checkpoint_name}*.pth"))
        assert len(model_checkpoints) == 1
        checkpoint_path = model_checkpoints[0]

    return str(checkpoint_path)


def get_results_table(
    noc_list,
    over_max_list,
    brs_type,
    dataset_name,
    mean_spc,
    elapsed_time,
    n_clicks=20,
    model_name=None,
):
    table_header = (
        f'|{"BRS Type":^13}|{"Dataset":^11}|'
        f'{"NoC@80%":^9}|{"NoC@85%":^9}|{"NoC@90%":^9}|'
        f'{">="+str(n_clicks)+"@85%":^9}|{">="+str(n_clicks)+"@90%":^9}|'
        f'{"SPC,s":^7}|{"Time":^9}|'
    )
    row_width = len(table_header)

    header = (
        _("Eval results for model: {model_name}\n").format(
            model_name=model_name
        )
        if model_name is not None
        else ""
    )
    header += "-" * row_width + "\n"
    header += table_header + "\n" + "-" * row_width

    eval_time = str(timedelta(seconds=int(elapsed_time)))
    table_row = f"|{brs_type:^13}|{dataset_name:^11}|"
    table_row += f"{noc_list[0]:^9.2f}|"
    table_row += f"{noc_list[1]:^9.2f}|" if len(noc_list) > 1 else f'{"?":^9}|'
    table_row += f"{noc_list[2]:^9.2f}|" if len(noc_list) > 2 else f'{"?":^9}|'
    table_row += (
        f"{over_max_list[1]:^9}|" if len(noc_list) > 1 else f'{"?":^9}|'
    )
    table_row += (
        f"{over_max_list[2]:^9}|" if len(noc_list) > 2 else f'{"?":^9}|'
    )
    table_row += f"{mean_spc:^7.3f}|{eval_time:^9}|"

    return header, table_row



import torch.nn as nn

from ritm_annotation.model.modifiers import LRMult
from ritm_annotation.utils.serialization import serialize

from .is_model import ISModel
from .modeling.hrnet_ocr import HighResolutionNet


class HRNetModel(ISModel):
    @serialize
    def __init__(
        self,
        width=48,
        ocr_width=256,
        small=False,
        backbone_lr_mult=0.1,
        norm_layer=nn.BatchNorm2d,
        **kwargs
    ):
        super().__init__(norm_layer=norm_layer, **kwargs)

        self.feature_extractor = HighResolutionNet(
            width=width,
            ocr_width=ocr_width,
            small=small,
            num_classes=1,
            norm_layer=norm_layer,
        )
        self.feature_extractor.apply(LRMult(backbone_lr_mult))
        if ocr_width > 0:
            self.feature_extractor.ocr_distri_head.apply(LRMult(1.0))
            self.feature_extractor.ocr_gather_head.apply(LRMult(1.0))
            self.feature_extractor.conv3x3_ocr.apply(LRMult(1.0))

    def backbone_forward(self, image, coord_features=None):
        net_outputs = self.feature_extractor(image, coord_features)

        return {"instances": net_outputs[0], "instances_aux": net_outputs[1]}



from typing import List

import torch

from ritm_annotation.inference.clicker import Click
from ritm_annotation.utils.misc import (
    clamp_bbox,
    expand_bbox,
    get_bbox_from_mask,
    get_bbox_iou,
)

from .base import BaseTransform


class ZoomIn(BaseTransform):
    def __init__(
        self,
        target_size=400,
        skip_clicks=1,
        expansion_ratio=1.4,
        min_crop_size=200,
        recompute_thresh_iou=0.5,
        prob_thresh=0.50,
    ):
        super().__init__()
        self.target_size = target_size
        self.min_crop_size = min_crop_size
        self.skip_clicks = skip_clicks
        self.expansion_ratio = expansion_ratio
        self.recompute_thresh_iou = recompute_thresh_iou
        self.prob_thresh = prob_thresh

        self._input_image_shape = None
        self._prev_probs = None
        self._object_roi = None
        self._roi_image = None

    def transform(self, image_nd, clicks_lists: List[List[Click]]):
        assert image_nd.shape[0] == 1 and len(clicks_lists) == 1
        self.image_changed = False

        clicks_list = clicks_lists[0]
        if len(clicks_list) <= self.skip_clicks:
            return image_nd, clicks_lists

        self._input_image_shape = image_nd.shape

        current_object_roi = None
        if self._prev_probs is not None:
            current_pred_mask = (self._prev_probs > self.prob_thresh)[0, 0]
            if current_pred_mask.sum() > 0:
                current_object_roi = get_object_roi(
                    current_pred_mask,
                    clicks_list,
                    self.expansion_ratio,
                    self.min_crop_size,
                )

        if current_object_roi is None:
            if self.skip_clicks >= 0:
                return image_nd, clicks_lists
            else:
                current_object_roi = (
                    0,
                    image_nd.shape[2] - 1,
                    0,
                    image_nd.shape[3] - 1,
                )

        update_object_roi = False
        if self._object_roi is None:
            update_object_roi = True
        elif not check_object_roi(self._object_roi, clicks_list):
            update_object_roi = True
        elif (
            get_bbox_iou(current_object_roi, self._object_roi)
            < self.recompute_thresh_iou
        ):
            update_object_roi = True

        if update_object_roi:
            self._object_roi = current_object_roi
            self.image_changed = True
        self._roi_image = get_roi_image_nd(
            image_nd, self._object_roi, self.target_size
        )

        tclicks_lists = [self._transform_clicks(clicks_list)]
        return self._roi_image.to(image_nd.device), tclicks_lists

    def inv_transform(self, prob_map):
        if self._object_roi is None:
            self._prev_probs = prob_map.cpu().numpy()
            return prob_map

        assert prob_map.shape[0] == 1
        rmin, rmax, cmin, cmax = self._object_roi
        prob_map = torch.nn.functional.interpolate(
            prob_map,
            size=(rmax - rmin + 1, cmax - cmin + 1),
            mode="bilinear",
            align_corners=True,
        )

        if self._prev_probs is not None:
            new_prob_map = torch.zeros(
                *self._prev_probs.shape,
                device=prob_map.device,
                dtype=prob_map.dtype
            )
            new_prob_map[:, :, rmin : rmax + 1, cmin : cmax + 1] = prob_map
        else:
            new_prob_map = prob_map

        self._prev_probs = new_prob_map.cpu().numpy()

        return new_prob_map

    def check_possible_recalculation(self):
        if (
            self._prev_probs is None
            or self._object_roi is not None
            or self.skip_clicks > 0
        ):
            return False

        pred_mask = (self._prev_probs > self.prob_thresh)[0, 0]
        if pred_mask.sum() > 0:
            possible_object_roi = get_object_roi(
                pred_mask, [], self.expansion_ratio, self.min_crop_size
            )
            image_roi = (
                0,
                self._input_image_shape[2] - 1,
                0,
                self._input_image_shape[3] - 1,
            )
            if get_bbox_iou(possible_object_roi, image_roi) < 0.50:
                return True
        return False

    def get_state(self):
        roi_image = (
            self._roi_image.cpu() if self._roi_image is not None else None
        )
        return (
            self._input_image_shape,
            self._object_roi,
            self._prev_probs,
            roi_image,
            self.image_changed,
        )

    def set_state(self, state):
        (
            self._input_image_shape,
            self._object_roi,
            self._prev_probs,
            self._roi_image,
            self.image_changed,
        ) = state

    def reset(self):
        self._input_image_shape = None
        self._object_roi = None
        self._prev_probs = None
        self._roi_image = None
        self.image_changed = False

    def _transform_clicks(self, clicks_list):
        if self._object_roi is None:
            return clicks_list

        rmin, rmax, cmin, cmax = self._object_roi
        crop_height, crop_width = self._roi_image.shape[2:]

        transformed_clicks = []
        for click in clicks_list:
            new_r = crop_height * (click.coords[0] - rmin) / (rmax - rmin + 1)
            new_c = crop_width * (click.coords[1] - cmin) / (cmax - cmin + 1)
            transformed_clicks.append(click.copy(coords=(new_r, new_c)))
        return transformed_clicks


def get_object_roi(pred_mask, clicks_list, expansion_ratio, min_crop_size):
    pred_mask = pred_mask.copy()

    for click in clicks_list:
        if click.is_positive:
            pred_mask[int(click.coords[0]), int(click.coords[1])] = 1

    bbox = get_bbox_from_mask(pred_mask)
    bbox = expand_bbox(bbox, expansion_ratio, min_crop_size)
    h, w = pred_mask.shape[0], pred_mask.shape[1]
    bbox = clamp_bbox(bbox, 0, h - 1, 0, w - 1)

    return bbox


def get_roi_image_nd(image_nd, object_roi, target_size):
    rmin, rmax, cmin, cmax = object_roi

    height = rmax - rmin + 1
    width = cmax - cmin + 1

    if isinstance(target_size, tuple):
        new_height, new_width = target_size
    else:
        scale = target_size / max(height, width)
        new_height = int(round(height * scale))
        new_width = int(round(width * scale))

    with torch.no_grad():
        roi_image_nd = image_nd[:, :, rmin : rmax + 1, cmin : cmax + 1]
        roi_image_nd = torch.nn.functional.interpolate(
            roi_image_nd,
            size=(new_height, new_width),
            mode="bilinear",
            align_corners=True,
        )

    return roi_image_nd


def check_object_roi(object_roi, clicks_list):
    for click in clicks_list:
        if click.is_positive:
            if (
                click.coords[0] < object_roi[0]
                or click.coords[0] >= object_roi[1]
            ):
                return False
            if (
                click.coords[1] < object_roi[2]
                or click.coords[1] >= object_roi[3]
            ):
                return False

    return True



import random

import cv2
import numpy as np
from albumentations import DualTransform, ImageOnlyTransform
from albumentations.augmentations import functional as F
from albumentations.augmentations.geometric import functional as FG
from albumentations.core.serialization import SERIALIZABLE_REGISTRY
from albumentations.core.transforms_interface import to_tuple

from ritm_annotation.utils.misc import (
    clamp_bbox,
    expand_bbox,
    get_bbox_from_mask,
    get_labels_with_sizes,
)


class UniformRandomResize(DualTransform):
    def __init__(
        self,
        scale_range=(0.9, 1.1),
        interpolation=cv2.INTER_LINEAR,
        always_apply=False,
        p=1,
    ):
        super().__init__(always_apply, p)
        self.scale_range = scale_range
        self.interpolation = interpolation

    def get_params_dependent_on_targets(self, params):
        scale = random.uniform(*self.scale_range)
        height = int(round(params["image"].shape[0] * scale))
        width = int(round(params["image"].shape[1] * scale))
        return {"new_height": height, "new_width": width}

    def apply(
        self,
        img,
        new_height=0,
        new_width=0,
        interpolation=cv2.INTER_LINEAR,
        **params
    ):
        return FG.resize(
            img,
            height=new_height,
            width=new_width,
            interpolation=interpolation,
        )

    def apply_to_keypoint(self, keypoint, new_height=0, new_width=0, **params):
        scale_x = new_width / params["cols"]
        scale_y = new_height / params["rows"]
        return F.keypoint_scale(keypoint, scale_x, scale_y)

    def get_transform_init_args_names(self):
        return "scale_range", "interpolation"

    @property
    def targets_as_params(self):
        return ["image"]


class ZoomIn(DualTransform):
    def __init__(
        self,
        height,
        width,
        bbox_jitter=0.1,
        expansion_ratio=1.4,
        min_crop_size=200,
        min_area=100,
        always_resize=False,
        always_apply=False,
        p=0.5,
    ):
        super(ZoomIn, self).__init__(always_apply, p)
        self.height = height
        self.width = width
        self.bbox_jitter = to_tuple(bbox_jitter)
        self.expansion_ratio = expansion_ratio
        self.min_crop_size = min_crop_size
        self.min_area = min_area
        self.always_resize = always_resize

    def apply(self, img, selected_object, bbox, **params):
        if selected_object is None:
            if self.always_resize:
                img = FG.resize(img, height=self.height, width=self.width)
            return img

        rmin, rmax, cmin, cmax = bbox
        img = img[rmin : rmax + 1, cmin : cmax + 1]
        img = FG.resize(img, height=self.height, width=self.width)

        return img

    def apply_to_mask(self, mask, selected_object, bbox, **params):
        if selected_object is None:
            if self.always_resize:
                mask = FG.resize(
                    mask,
                    height=self.height,
                    width=self.width,
                    interpolation=cv2.INTER_NEAREST,
                )
            return mask

        rmin, rmax, cmin, cmax = bbox
        mask = mask[rmin : rmax + 1, cmin : cmax + 1]
        if isinstance(selected_object, tuple):
            layer_indx, mask_id = selected_object
            obj_mask = mask[:, :, layer_indx] == mask_id
            new_mask = np.zeros_like(mask)
            new_mask[:, :, layer_indx][obj_mask] = mask_id
        else:
            obj_mask = mask == selected_object
            new_mask = mask.copy()
            new_mask[np.logical_not(obj_mask)] = 0

        new_mask = FG.resize(
            new_mask,
            height=self.height,
            width=self.width,
            interpolation=cv2.INTER_NEAREST,
        )
        return new_mask

    def get_params_dependent_on_targets(self, params):
        instances = params["mask"]

        is_mask_layer = len(instances.shape) > 2
        candidates = []
        if is_mask_layer:
            for layer_indx in range(instances.shape[2]):
                labels, areas = get_labels_with_sizes(
                    instances[:, :, layer_indx]
                )
                candidates.extend(
                    [
                        (layer_indx, obj_id)
                        for obj_id, area in zip(labels, areas)
                        if area > self.min_area
                    ]
                )
        else:
            labels, areas = get_labels_with_sizes(instances)
            candidates = [
                obj_id
                for obj_id, area in zip(labels, areas)
                if area > self.min_area
            ]

        selected_object = None
        bbox = None
        if candidates:
            selected_object = random.choice(candidates)
            if is_mask_layer:
                layer_indx, mask_id = selected_object
                obj_mask = instances[:, :, layer_indx] == mask_id
            else:
                obj_mask = instances == selected_object

            bbox = get_bbox_from_mask(obj_mask)

            if isinstance(self.expansion_ratio, tuple):
                expansion_ratio = random.uniform(*self.expansion_ratio)
            else:
                expansion_ratio = self.expansion_ratio

            bbox = expand_bbox(bbox, expansion_ratio, self.min_crop_size)
            bbox = self._jitter_bbox(bbox)
            bbox = clamp_bbox(
                bbox, 0, obj_mask.shape[0] - 1, 0, obj_mask.shape[1] - 1
            )

        return {"selected_object": selected_object, "bbox": bbox}

    def _jitter_bbox(self, bbox):
        rmin, rmax, cmin, cmax = bbox
        height = rmax - rmin + 1
        width = cmax - cmin + 1
        rmin = int(rmin + random.uniform(*self.bbox_jitter) * height)
        rmax = int(rmax + random.uniform(*self.bbox_jitter) * height)
        cmin = int(cmin + random.uniform(*self.bbox_jitter) * width)
        cmax = int(cmax + random.uniform(*self.bbox_jitter) * width)

        return rmin, rmax, cmin, cmax

    def apply_to_bbox(self, bbox, **params):
        raise NotImplementedError

    def apply_to_keypoint(self, keypoint, **params):
        raise NotImplementedError

    @property
    def targets_as_params(self):
        return ["mask"]

    def get_transform_init_args_names(self):
        return (
            "height",
            "width",
            "bbox_jitter",
            "expansion_ratio",
            "min_crop_size",
            "min_area",
            "always_resize",
        )


def remove_image_only_transforms(sdict):
    if "transforms" not in sdict:
        return sdict

    keep_transforms = []
    for tdict in sdict["transforms"]:
        cls = SERIALIZABLE_REGISTRY[tdict["__class_fullname__"]]
        if "transforms" in tdict:
            keep_transforms.append(remove_image_only_transforms(tdict))
        elif not issubclass(cls, ImageOnlyTransform):
            keep_transforms.append(tdict)
    sdict["transforms"] = keep_transforms

    return sdict



import json
import random
from collections import defaultdict
from pathlib import Path

import cv2
import numpy as np

from ritm_annotation.data.base import ISDataset
from ritm_annotation.data.sample import DSample


class LvisDataset(ISDataset):
    def __init__(
        self,
        dataset_path,
        split="train",
        max_overlap_ratio=0.5,
        dry_run=False,
        **kwargs,
    ):
        super(LvisDataset, self).__init__(**kwargs)
        dataset_path = Path(dataset_path)
        train_categories_path = dataset_path / "train_categories.json"
        self._train_path = dataset_path / "train"
        self._val_path = dataset_path / "val"

        self.split = split
        self.max_overlap_ratio = max_overlap_ratio

        if not dry_run:
            with open(
                dataset_path / split / f"lvis_{self.split}.json", "r"
            ) as f:
                json_annotation = json.loads(f.read())
        else:
            json_annotation = dict(annotations=[], images=[])

        self.annotations = defaultdict(list)
        for x in json_annotation["annotations"]:
            self.annotations[x["image_id"]].append(x)

        if not train_categories_path.exists():
            self.generate_train_categories(
                dataset_path, train_categories_path, dry_run=dry_run
            )
        self.dataset_samples = [
            x
            for x in json_annotation["images"]
            if len(self.annotations[x["id"]]) > 0
        ]

    def get_sample(self, index) -> DSample:
        image_info = self.dataset_samples[index]
        image_id, image_url = image_info["id"], image_info["coco_url"]
        image_filename = image_url.split("/")[-1]
        image_annotations = self.annotations[image_id]
        random.shuffle(image_annotations)

        # LVISv1 splits do not match older LVIS splits (some images in val may come from COCO train2017)  # noqa: E501
        if "train2017" in image_url:
            image_path = self._train_path / "images" / image_filename
        else:
            image_path = self._val_path / "images" / image_filename
        image = cv2.imread(str(image_path))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        instances_mask = None
        instances_area: dict[int, int] = defaultdict(int)
        objects_ids = []
        for indx, obj_annotation in enumerate(image_annotations):
            mask = self.get_mask_from_polygon(obj_annotation, image)
            object_mask = mask > 0
            object_area = object_mask.sum()

            if instances_mask is None:
                instances_mask = np.zeros_like(object_mask, dtype=np.int32)

            overlap_ids = np.bincount(instances_mask[object_mask].flatten())
            overlap_areas = [
                overlap_area / instances_area[inst_id]
                for inst_id, overlap_area in enumerate(overlap_ids)
                if overlap_area > 0 and inst_id > 0
            ]
            overlap_ratio = (
                np.logical_and(object_mask, instances_mask > 0).sum()
                / object_area
            )
            if overlap_areas:
                overlap_ratio = max(overlap_ratio, max(overlap_areas))
            if overlap_ratio > self.max_overlap_ratio:
                continue

            instance_id = indx + 1
            instances_mask[object_mask] = instance_id
            instances_area[instance_id] = object_area
            objects_ids.append(instance_id)

        return DSample(image, instances_mask, objects_ids=objects_ids)

    @staticmethod
    def get_mask_from_polygon(annotation, image):
        mask = np.zeros(image.shape[:2], dtype=np.int32)
        for contour_points in annotation["segmentation"]:
            contour_points = np.array(contour_points).reshape((-1, 2))
            contour_points = np.round(contour_points).astype(np.int32)[
                np.newaxis, :
            ]
            cv2.fillPoly(mask, contour_points, 1)

        return mask

    @staticmethod
    def generate_train_categories(
        dataset_path, train_categories_path, dry_run=False
    ):
        if not dry_run:
            with open(dataset_path / "train/lvis_train.json", "r") as f:
                annotation = json.load(f)
        else:
            annotation = dict(categories=[])

        with open(train_categories_path, "w") as f:
            json.dump(annotation["categories"], f, indent=1)
